Package: adjustr
Encoding: UTF-8
Type: Package
Title: Stan Model Adjustments and Sensitivity Analyses using Importance
        Sampling
Version: 0.2.0
Authors@R: person("Cory", "McCartan", email = "mccartan@psu.edu",
  role = c("aut", "cre", "cph"))
Description: Assess the sensitivity of a Bayesian model (fitted using 'Stan'
    via 'rstan', 'brms', or 'cmdstanr') to the specification of its likelihood
    and priors. Users provide a series of alternate sampling specifications, and
    the package uses Pareto-smoothed importance sampling (PSIS) to estimate
    posterior quantities of interest under each specification, without needing
    to refit the model. Methods are based on Vehtari, Simpson, Gelman, Yao,
    and Gabry (2024) <doi:10.48550/arXiv.1507.02646>.
License: MIT + file LICENSE
Depends: R (>= 3.6.0), dplyr (>= 1.0.0)
Imports: rlang, tidyselect, rstan, loo
Suggests: ggplot2, extraDistr, tidyr, testthat, covr, knitr, rmarkdown
URL: https://corymccartan.com/adjustr/
BugReports: https://github.com/CoryMcCartan/adjustr/issues
LazyData: true
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-22 22:09:00 UTC; cmccartan
Author: Cory McCartan [aut, cre, cph]
Maintainer: Cory McCartan <mccartan@psu.edu>
Repository: CRAN
Date/Publication: 2026-05-29 09:20:02 UTC
Built: R 4.5.2; ; 2026-05-29 10:23:53 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(arrange,adjustr_spec)
S3method(as.data.frame,adjustr_spec)
S3method(dplyr::filter,adjustr_spec)
S3method(length,adjustr_spec)
S3method(print,adjustr_spec)
S3method(pull,adjustr_weighted)
S3method(rename,adjustr_spec)
S3method(select,adjustr_spec)
S3method(slice,adjustr_spec)
S3method(stats::quantile,weighted.ecdf)
S3method(summarise,adjustr_weighted)
S3method(summarize,adjustr_weighted)
export(adjust_weights)
export(extract_samp_stmts)
export(get_resampling_idxs)
export(make_spec)
export(spec_plot)
import(dplyr)
import(rlang)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# adjustr <a href="https://corymccartan.com/adjustr/"><img src="man/figures/logo.png" align="right" height="138" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/CoryMcCartan/adjustr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/CoryMcCartan/adjustr/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/CoryMcCartan/adjustr/branch/main/graph/badge.svg)](https://app.codecov.io/gh/CoryMcCartan/adjustr?branch=main)
<!-- badges: end -->

Sensitivity analysis is a critical component of a good modeling
workflow. Yet as the number and power of Bayesian computational tools
has increased, the options for sensitivity analysis have remained
largely the same: compute importance sampling weights manually, or fit a
large number of similar models, dramatically increasing computation
time. Neither option is satisfactory for most applied modeling.

**adjustr** is an R package which aims to make sensitivity analysis
faster and easier, and works with Bayesian models fitted with
[Stan](https://mc-stan.org). Users provide a series of alternate
sampling specifications, and the package uses Pareto-smoothed importance
sampling to estimate the posterior under each specification. The package
also provides functions to summarize and plot how posterior quantities
change across specifications.

The package provides a simple interface that makes it as easy as
possible for researchers to try out various adjustments to their Stan
models, without needing to write any specific Stan code or even
recompile or rerun their model.

The package works by parsing Stan model code, so everything works best
if the model was written by the user. Models made using **brms** may in
principle be used as well. Models made using **rstanarm** are
constructed using more complex model templates, and cannot be used.

## Installation

Install from CRAN with:

``` r
install.packages("adjustr")
```

Or install the development version from GitHub:

``` r
# install.packages("remotes")
remotes::install_github("CoryMcCartan/adjustr")
```

## Getting Started

The basic **adjustr** workflow is as follows:

1.  Use `make_spec()` to specify the set of alternative model
    specifications you’d like to fit.

2.  Use `adjust_weights()` to calculate importance sampling weights
    which approximate the posterior of each alternative specification.

3.  Use `summarize()` and `spec_plot()` to examine posterior quantities
    of interest for each alternative specification, in order to assess
    the sensitivity of the underlying model.

To illustrate, the package lets us do the following:

``` r
extract_samp_stmts(eightschools_m)
#> Sampling statements for model 2c8d1d8a30137533422c438f23b83428:
#>   parameter   eta ~ std_normal()
#>   data        y ~ normal(theta, sigma)

make_spec(eta ~ student_t(0, 1, df), df=1:10) %>%
    adjust_weights(eightschools_m) %>%
    summarize(wasserstein(mu))
#> # A tibble: 11 x 5
#>       df .samp                     .weights      .pareto_k `wasserstein(mu)`
#>    <int> <chr>                     <list>            <dbl>             <dbl>
#>  1     1 eta ~ student_t(df, 0, 1) <dbl [1,000]>     1.02              0.928
#>  2     2 eta ~ student_t(df, 0, 1) <dbl [1,000]>     1.03              0.736
#>  3     3 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.915             0.534
#>  4     4 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.856             0.411
#>  5     5 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.826             0.341
#>  6     6 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.803             0.275
#>  7     7 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.782             0.234
#>  8     8 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.753             0.195
#>  9     9 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.736             0.166
#> 10    10 eta ~ student_t(df, 0, 1) <dbl [1,000]>     0.721             0.151
#> 11    NA <original model>          <dbl [1,000]>  -Inf                 0
```

The tutorial
[vignette](https://corymccartan.com/adjustr/articles/adjustr.html) walks
through a full sensitivity analysis for this 8-schools example. Smaller
examples are also included in the package
[documentation](https://corymccartan.com/adjustr/reference/index.html).

## References

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., & Gabry, J. (2024).
Pareto smoothed importance sampling. *Journal of Machine Learning
Research*, 25(72), 1–58. <https://doi.org/10.48550/arXiv.1507.02646>

# Token types used by the Stan tokenizer
# Order matters: longer/more-specific patterns first; word boundaries prevent
# partial matches (e.g. "in" inside "int").
TOKEN_TYPES = c(
    BLOCK_KW   = "(?:transformed\\s+data|transformed\\s+parameters|generated\\s+quantities|functions|parameters|data|model)\\b",
    TYPE_KW    = paste0("(?:complex_row_vector|complex_vector|complex_matrix|",
                        "cholesky_factor_corr|cholesky_factor_cov|",
                        "column_stochastic_matrix|row_stochastic_matrix|",
                        "sum_to_zero_vector|sum_to_zero_matrix|",
                        "positive_ordered|row_vector|cov_matrix|corr_matrix|",
                        "unit_vector|",
                        "ordered|simplex|complex|vector|matrix|array|tuple|void|",
                        "real|int)\\b"),
    KEYWORD    = "(?:continue|profile|jacobian|return|reject|target|break|while|print|else|for|if|in)\\b",
    NUMBER     = "(?:[0-9]+\\.?[0-9]*(?:[eE][+-]?[0-9]+)?|\\.?[0-9]+(?:[eE][+-]?[0-9]+)?)",
    STRING     = '"[^"]*"',
    IDENTIFIER = "[a-zA-Z_][a-zA-Z0-9_]*",
    PLUSASSIGN  = "\\+=",
    TILDE      = "~",
    LBRACE     = "\\{",
    RBRACE     = "\\}",
    LBRACK     = "\\[",
    RBRACK     = "\\]",
    LPAREN     = "\\(",
    RPAREN     = "\\)",
    LANGLE     = "<",
    RANGLE     = ">",
    SEMICOLON  = ";",
    COMMA      = ",",
    BAR        = "\\|",
    COLON      = ":",
    OP         = "(?:\\+|-|\\*|/|%|\\^|\\.\\*|\\./|\\.\\^|==|!=|<=|>=|&&|\\|\\||!|'|\\.)"
)

# Combined pattern: try each token type in order
TOKEN_RE = paste0("(", TOKEN_TYPES, ")", collapse="|")

# Tokenize Stan code into a data frame with columns: type, value
tokenize_stan = function(code) {
    # Strip comments
    code = gsub("//[^\n]*", "", code)
    code = gsub("/\\*[^*]*\\*+(?:[^/*][^*]*\\*+)*/", "", code, perl=TRUE)

    # Find all tokens
    matches = regmatches(code, gregexpr(TOKEN_RE, code, perl=TRUE))[[1]]
    if (length(matches) == 0) {
        return(data.frame(type = character(0), value = character(0),
                          stringsAsFactors = FALSE))
    }
    # Match each token against each type pattern to determine its type
    type_idx = vapply(matches, function(m) {
        for (i in seq_along(TOKEN_TYPES)) {
            if (grepl(paste0("^", TOKEN_TYPES[i], "$"), m, perl=TRUE)) return(i)
        }
        NA_integer_
    }, integer(1))
    types = names(TOKEN_TYPES)[type_idx]
    values = matches

    # Normalize whitespace in block keywords
    values = gsub("\\s+", " ", values)

    data.frame(type = types, value = values, stringsAsFactors = FALSE)
}


# Extract top-level blocks from token stream.
# Returns a named list of token data frames, one per block.
extract_blocks = function(tokens) {
    block_kw_idx = which(tokens$type == "BLOCK_KW")
    if (length(block_kw_idx) == 0) return(list())

    blocks = list()
    for (i in seq_along(block_kw_idx)) {
        bk_idx = block_kw_idx[i]
        block_name = tokens$value[bk_idx]

        # Find the opening brace
        open_idx = bk_idx + 1
        if (open_idx > nrow(tokens) || tokens$value[open_idx] != "{") next

        # Walk forward counting braces to find matching close
        depth = 0
        close_idx = NA
        for (j in open_idx:nrow(tokens)) {
            if (tokens$value[j] == "{") depth = depth + 1
            else if (tokens$value[j] == "}") depth = depth - 1
            if (depth == 0) { close_idx = j; break }
        }
        if (is.na(close_idx)) next

        # Extract tokens inside the block (excluding outer braces)
        if (close_idx > open_idx + 1) {
            block_tokens = tokens[(open_idx + 1):(close_idx - 1), , drop = FALSE]
            rownames(block_tokens) = NULL
        } else {
            block_tokens = data.frame(type = character(0), value = character(0),
                                      stringsAsFactors = FALSE)
        }
        blocks[[block_name]] = block_tokens
    }

    blocks
}


# Parse variable declarations from a block's token stream.
# Returns a character vector of variable names.
# Handles both old-style `real y` and new-style `array[N] real y`.
parse_declarations = function(tokens) {
    if (nrow(tokens) == 0) return(character(0))

    vars = character(0)
    i = 1
    n = nrow(tokens)
    depth = 0  # brace depth -- only parse at top level

    while (i <= n) {
        # Track brace depth
        if (tokens$value[i] == "{") { depth = depth + 1; i = i + 1; next }
        if (tokens$value[i] == "}") { depth = depth - 1; i = i + 1; next }

        # Only parse declarations at top level (depth 0)
        if (depth != 0) { i = i + 1; next }

        # Check for type keyword (possibly preceded by array dims)
        start_i = i
        is_decl = FALSE

        # Optional: array[...] prefix
        if (tokens$value[i] == "array" && tokens$type[i] == "TYPE_KW") {
            # Skip array[dims]
            i = i + 1
            i = skip_brackets(tokens, i, "[", "]")
        }

        # Optional: tuple(...) type
        if (i <= n && tokens$value[i] == "tuple" && tokens$type[i] == "TYPE_KW") {
            i = i + 1
            i = skip_brackets(tokens, i, "(", ")")
            is_decl = TRUE
        } else if (i <= n && tokens$type[i] == "TYPE_KW") {
            # Standard type keyword
            i = i + 1
            is_decl = TRUE
        }

        if (!is_decl) {
            # Not a declaration; skip to next semicolon at this depth
            i = start_i
            i = skip_to_semicolon(tokens, i, depth)
            next
        }

        # Optional: type constraints <lower=..., upper=...> or <...>
        if (i <= n && tokens$value[i] == "<") {
            i = skip_brackets(tokens, i, "<", ">")
        }

        # Optional: dimensions [...]
        if (i <= n && tokens$value[i] == "[") {
            i = skip_brackets(tokens, i, "[", "]")
        }

        # Now we should be at the variable name (an identifier)
        if (i <= n && tokens$type[i] == "IDENTIFIER") {
            vars = c(vars, tokens$value[i])
            i = i + 1
            # Old-style trailing dimensions: real y[J];
            if (i <= n && tokens$value[i] == "[") {
                i = skip_brackets(tokens, i, "[", "]")
            }
        }

        # Skip rest of declaration (possible assignment, more declarators) to semicolon
        i = skip_to_semicolon(tokens, i, depth)
    }

    vars
}


# Parse sampling statements from a block's token stream.
# Returns a list of R formulas.
# Only extracts statements at brace depth 0 (top-level within the block).
parse_sampling = function(tokens) {
    if (nrow(tokens) == 0) return(list())

    samps = list()
    i = 1
    n = nrow(tokens)
    depth = 0

    while (i <= n) {
        if (tokens$value[i] == "{") { depth = depth + 1; i = i + 1; next }
        if (tokens$value[i] == "}") { depth = depth - 1; i = i + 1; next }

        if (depth != 0) { i = i + 1; next }

        # Pattern 1: expr ~ dist(args);
        # Look for TILDE at top level
        tilde_pos = find_token_in_stmt(tokens, i, "TILDE")
        semi_pos = find_token_in_stmt(tokens, i, "SEMICOLON")

        if (!is.na(tilde_pos) && !is.na(semi_pos) && tilde_pos < semi_pos) {
            samp = parse_tilde_stmt(tokens, i, tilde_pos, semi_pos)
            if (!is.null(samp)) samps = c(samps, list(samp))
            i = semi_pos + 1
            next
        }

        # Pattern 2: target += dist_lpdf(expr | args);  or  dist_lpmf(...)
        if (!is.na(semi_pos) && tokens$value[i] == "target" &&
            i + 1 <= n && tokens$value[i + 1] == "+=") {
            samp = parse_target_stmt(tokens, i + 2, semi_pos)
            if (!is.null(samp)) samps = c(samps, list(samp))
            i = semi_pos + 1
            next
        }

        # Not a sampling statement; skip to next semicolon
        if (!is.na(semi_pos)) {
            i = semi_pos + 1
        } else {
            i = i + 1
        }
    }

    samps
}


# Parse `lhs ~ dist(args)` into an R formula
parse_tilde_stmt = function(tokens, start, tilde_pos, semi_pos) {
    # LHS: tokens from start to tilde_pos - 1
    lhs_str = paste(tokens$value[start:(tilde_pos - 1)], collapse = "")
    # Strip any trailing indexing from LHS for the formula variable name
    lhs_var = extract_base_var(lhs_str)

    # RHS: tokens from tilde_pos + 1 to semi_pos - 1
    rhs_tokens = tokens[(tilde_pos + 1):(semi_pos - 1), , drop = FALSE]

    # Handle optional truncation T[...] at the end -- just strip it
    rhs_tokens = strip_truncation(rhs_tokens)

    rhs_str = stan_expr_to_r(rhs_tokens)

    make_formula(lhs_var, rhs_str)
}


# Parse `target += dist_lpdf(expr | args)` into an R formula
# Returns NULL if not a recognized distribution call
parse_target_stmt = function(tokens, start, semi_pos) {
    if (start >= semi_pos) return(NULL)

    expr_tokens = tokens[start:(semi_pos - 1), , drop = FALSE]
    # Look for a function call with _lpdf, _lpmf, _lupdf, _lupmf suffix
    # at the top level of the expression
    if (nrow(expr_tokens) < 1) return(NULL)

    # The expression might be just `dist_lpdf(x | args)` or could have
    # additional terms. We only handle the simple case.
    func_name = expr_tokens$value[1]
    if (expr_tokens$type[1] != "IDENTIFIER") return(NULL)

    suffixes = c("_lpdf", "_lpmf", "_lupdf", "_lupmf", "_lpmf", "_log")
    matched_suffix = NULL
    for (sfx in suffixes) {
        if (endsWith(func_name, sfx)) {
            matched_suffix = sfx
            break
        }
    }
    if (is.null(matched_suffix)) return(NULL)

    dist_name = sub(paste0(matched_suffix, "$"), "", func_name)

    # Expect ( after function name
    if (nrow(expr_tokens) < 3 || expr_tokens$value[2] != "(") return(NULL)

    # Find the matching )
    close_paren = find_matching(expr_tokens, 2, "(", ")")
    if (is.na(close_paren)) return(NULL)

    # Tokens inside parens
    inner = expr_tokens[3:(close_paren - 1), , drop = FALSE]
    if (nrow(inner) == 0) return(NULL)

    # Find the | separator (at top level within these parens)
    bar_pos = find_bar_in_tokens(inner)

    if (!is.na(bar_pos)) {
        # New syntax: dist_lpdf(x | args)
        lhs_tokens = inner[1:(bar_pos - 1), , drop = FALSE]
        if (bar_pos < nrow(inner)) {
            args_tokens = inner[(bar_pos + 1):nrow(inner), , drop = FALSE]
        } else {
            args_tokens = data.frame(type = character(0), value = character(0),
                                     stringsAsFactors = FALSE)
        }
    } else if (matched_suffix == "_log") {
        # Old syntax: dist_log(x, args) -- first arg is the variate
        # Find first top-level comma
        comma_pos = find_comma_in_tokens(inner)
        if (is.na(comma_pos)) {
            lhs_tokens = inner
            args_tokens = data.frame(type = character(0), value = character(0),
                                     stringsAsFactors = FALSE)
        } else {
            lhs_tokens = inner[1:(comma_pos - 1), , drop = FALSE]
            args_tokens = inner[(comma_pos + 1):nrow(inner), , drop = FALSE]
        }
    } else {
        # No | found in lpdf/lpmf -- might be single-arg distribution
        lhs_tokens = inner
        args_tokens = data.frame(type = character(0), value = character(0),
                                 stringsAsFactors = FALSE)
    }

    lhs_str = paste(lhs_tokens$value, collapse = "")
    lhs_var = extract_base_var(lhs_str)

    if (nrow(args_tokens) > 0) {
        rhs_str = paste0(dist_name, "(", stan_expr_to_r(args_tokens), ")")
    } else {
        rhs_str = paste0(dist_name, "()")
    }

    make_formula(lhs_var, rhs_str)
}


# --- Helper functions ---

# Skip balanced brackets starting at position i.
# Returns position after the closing bracket.
skip_brackets = function(tokens, i, open, close) {
    n = nrow(tokens)
    if (i > n || tokens$value[i] != open) return(i)
    depth = 0
    while (i <= n) {
        if (tokens$value[i] == open) depth = depth + 1
        else if (tokens$value[i] == close) depth = depth - 1
        i = i + 1
        if (depth == 0) break
    }
    i
}

# Skip to semicolon at the given brace depth, return position after it
skip_to_semicolon = function(tokens, i, target_depth) {
    n = nrow(tokens)
    depth = target_depth
    while (i <= n) {
        if (tokens$value[i] == "{") depth = depth + 1
        else if (tokens$value[i] == "}") depth = depth - 1
        else if (tokens$value[i] == ";" && depth == target_depth) return(i + 1)
        i = i + 1
    }
    i
}

# Find a token of given type in the current statement (up to next ; at depth 0)
# Only matches at paren/bracket depth 0 within the statement
find_token_in_stmt = function(tokens, start, token_type) {
    n = nrow(tokens)
    depth = 0  # paren/bracket depth
    for (i in start:n) {
        v = tokens$value[i]
        if (v %in% c("(", "[")) depth = depth + 1
        else if (v %in% c(")", "]")) depth = depth - 1
        if (depth == 0 && tokens$type[i] == token_type) return(i)
        if (depth == 0 && v == ";" ) return(if (token_type == "SEMICOLON") i else NA)
        if (v == "{" || v == "}") return(NA)  # hit nested block
    }
    NA
}

# Find matching close bracket
find_matching = function(tokens, open_pos, open, close) {
    n = nrow(tokens)
    depth = 0
    for (i in open_pos:n) {
        if (tokens$value[i] == open) depth = depth + 1
        else if (tokens$value[i] == close) depth = depth - 1
        if (depth == 0) return(i)
    }
    NA
}

# Find | at top level in a token data frame
find_bar_in_tokens = function(tokens) {
    depth = 0
    for (i in seq_len(nrow(tokens))) {
        v = tokens$value[i]
        if (v %in% c("(", "[")) depth = depth + 1
        else if (v %in% c(")", "]")) depth = depth - 1
        else if (v == "|" && depth == 0) return(i)
    }
    NA
}

# Find first comma at top level in a token data frame
find_comma_in_tokens = function(tokens) {
    depth = 0
    for (i in seq_len(nrow(tokens))) {
        v = tokens$value[i]
        if (v %in% c("(", "[")) depth = depth + 1
        else if (v %in% c(")", "]")) depth = depth - 1
        else if (v == "," && depth == 0) return(i)
    }
    NA
}

# Extract base variable name from an expression like "y[n]" or "Y[n, m]"
extract_base_var = function(expr_str) {
    regmatches(expr_str, regexpr("^[a-zA-Z_][a-zA-Z0-9_]*", expr_str))
}

# Convert Stan expression tokens to an R-compatible expression string.
# Handles Stan-to-R operator differences.
stan_expr_to_r = function(tokens) {
    values = tokens$value
    # Stan integer division is %/% in R (Stan uses %/% too, but sometimes \)
    paste(values, collapse = "")
}

# Strip trailing T[...] truncation from RHS tokens
strip_truncation = function(tokens) {
    n = nrow(tokens)
    if (n < 1) return(tokens)

    # Check if last token is ] and there's a T before a [
    if (tokens$value[n] != "]") return(tokens)

    # Walk back to find the matching [
    depth = 0
    for (i in n:1) {
        if (tokens$value[i] == "]") depth = depth + 1
        else if (tokens$value[i] == "[") depth = depth - 1
        if (depth == 0) {
            # Check if preceded by T
            if (i > 1 && tokens$value[i - 1] == "T" && tokens$type[i - 1] == "IDENTIFIER") {
                return(tokens[1:(i - 2), , drop = FALSE])
            }
            return(tokens)
        }
    }
    tokens
}

# Create a formula from LHS and RHS strings, or return NULL on failure
make_formula = function(lhs, rhs) {
    stmt = paste(lhs, "~", rhs)
    tryCatch(
        stats::as.formula(stmt, env = rlang::empty_env()),
        error = function(e) NULL
    )
}


# Parse Stan `model_code` into a list with two elements:
#   `vars` — named character vector mapping variable names to their block
#   `samp` — list of sampling statements as R formulas
parse_model = function(model_code) {
    tokens = tokenize_stan(model_code)
    blocks = extract_blocks(tokens)

    if (length(blocks) == 0) return(list(vars = character(0), samps = list()))

    # Extract variable declarations from each block
    vars = character(0)
    for (block_name in names(blocks)) {
        block_tokens = blocks[[block_name]]
        block_vars = parse_declarations(block_tokens)
        if (length(block_vars) > 0) {
            new_vars = rep(block_name, length(block_vars))
            names(new_vars) = block_vars
            vars = c(vars, new_vars)
        }
    }

    # Extract sampling statements from model block
    samps = list()
    if ("model" %in% names(blocks)) {
        samps = parse_sampling(blocks[["model"]])
    }

    # Add implicit uniform priors for parameters without sampling statements
    parameters = names(vars)[vars == "parameters"]
    sampled_pars = vapply(samps, function(s) deparse(rlang::f_lhs(s)), "")
    uniform_pars = setdiff(parameters, sampled_pars)
    if (length(uniform_pars) > 0) {
        uniform_samp = lapply(uniform_pars, function(p) {
            stats::as.formula(paste0(p, " ~ uniform(-1e100, 1e100)"),
                              env = rlang::empty_env())
        })
    } else {
        uniform_samp = NULL
    }

    list(vars = vars, samp = c(samps, uniform_samp))
}


# Take a list of provided sampling formulas and return a matching list of
# sampling statements from a reference list
match_sampling_stmts = function(new_samp, ref_samp) {
    ref_vars = vapply(ref_samp, function(s) deparse(rlang::f_lhs(s)), "")
    new_vars = vapply(new_samp, function(s) deparse(rlang::f_lhs(s)), "")
    indices = match(new_vars, ref_vars)
    # check that every prior was matched
    if (any(is.na(indices))) {
        stop("No matching sampling statement found for ",
             new_samp[which.max(is.na(indices))],
             "\n  Check sampling statements and ensure that model data ",
             "has been provided.")
    }
    ref_samp[indices]
}

# Extract a list of variables from a sampling statement
# R versions of mathematical operators must be used
get_stmt_vars = function(stmt) {
    get_ast = function(x) {
        xl = as.list(x)
        lapply(xl, function(el) if (rlang::is_call(el)) get_ast(el) else el)
    }
    if (!rlang::is_call(rlang::f_rhs(stmt)))
        stop("Sampling statment ", format(stmt),
             " does not contain a distribution on the right-hand side.")
    # pull out variables from RHS
    ops = c("`+`", "`-`", "`*`", "`/`", "`^`", "`%*%`", "`%%`")
    rhs_vars = rlang::call_args(rlang::f_rhs(stmt))
    rhs_vars = unlist(get_ast(rhs_vars))
    rhs_vars = rhs_vars[!vapply(rhs_vars, is.numeric, logical(1))]
    rhs_vars = as.character(rhs_vars)
    rhs_vars = rhs_vars[!rhs_vars %in% ops]
    c(deparse(rlang::f_lhs(stmt)), rhs_vars)
}

context("Stan model parsing")

test_that("Empty model handled correctly", {
    parsed = parse_model("")
    expect_equal(length(parsed$vars), 0)
    expect_equal(length(parsed$samps), 0)
})

test_that("Correct parsed variables", {
    correct_vars = c(J= "data", y= "data", sigma="data", mu="parameters",
                     tau="parameters", eta="parameters",
                     theta="transformed parameters")
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    expect_equal(parsed$vars, correct_vars)
})

test_that("Correct parsed sampling statements", {
    correct_samp = list(eta ~ std_normal(), y ~ normal(theta, sigma),
                        mu ~ uniform(-1e+100, 1e+100), tau ~ uniform(-1e+100, 1e+100))
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    expect_equal(parsed$samp, correct_samp)
})

test_that("Provided sampling statements can be matched to model", {
    model_samp = list(eta ~ std_normal(), y ~ normal(theta, sigma))
    prov_samp = list(eta ~ exponential(5))
    matched = match_sampling_stmts(prov_samp, model_samp)
    expect_false(identical(matched, prov_samp))
    expect_length(matched, 1)
    expect_equal(matched[[1]], model_samp[[1]])
})

test_that("Extra sampling statements not in model throw an error", {
    model_samp = list(eta ~ std_normal(), y ~ normal(theta, sigma))
    prov_samp = list(eta ~ exponential(5), x ~ normal(theta, sigma))
    expect_error(match_sampling_stmts(prov_samp, model_samp),
                 "No matching sampling statement found for x ~ normal\\(theta, sigma\\)")
})

test_that("Variables are correctly extracted from sampling statements", {
    expect_equal(get_stmt_vars(y ~ normal(theta %*% eta/2, 2%%4 + sigma)),
                 c("y", "theta", "eta", "sigma"))
    expect_error(get_stmt_vars(y ~ .), "y ~ \\. does not")
})

# --- Tests for new token-based parser ---

test_that("New array syntax is parsed correctly", {
    code = "
    data {
        array[N] real y;
        array[N, K] int x;
    }
    parameters {
        array[3] vector[N] beta;
    }
    model {
        y ~ normal(0, 1);
    }"
    parsed = parse_model(code)
    expect_equal(parsed$vars[c("y", "x", "beta")],
                 c(y = "data", x = "data", beta = "parameters"))
})

test_that("Old-style array syntax is still parsed correctly", {
    code = "
    data {
        int<lower=0> N;
        real y[N];
        real<lower=0> sigma[N];
    }
    parameters {
        real mu;
    }
    model {
        y ~ normal(mu, sigma);
    }"
    parsed = parse_model(code)
    expect_equal(parsed$vars,
                 c(N = "data", y = "data", sigma = "data", mu = "parameters"))
})

test_that("target += with lpdf/lpmf is parsed correctly", {
    code = "
    data {
        int<lower=1> N;
        array[N] int Y;
    }
    parameters {
        real mu;
    }
    model {
        target += normal_lpdf(mu | 0, 10);
        target += poisson_lpmf(Y | mu);
    }"
    parsed = parse_model(code)
    expect_length(parsed$samp, 2)
    expect_equal(parsed$samp[[1]], mu ~ normal(0, 10))
    expect_equal(parsed$samp[[2]], Y ~ poisson(mu))
})

test_that("Statements inside control flow are skipped", {
    code = "
    data {
        int<lower=1> N;
        array[N] real y;
    }
    parameters {
        real mu;
        real<lower=0> sigma;
    }
    model {
        mu ~ normal(0, 10);
        if (N > 0) {
            y ~ normal(mu, sigma);
        }
    }"
    parsed = parse_model(code)
    # Only the top-level statement should be found
    sampled = vapply(parsed$samp, function(s) deparse(rlang::f_lhs(s)), "")
    expect_true("mu" %in% sampled)
    expect_false("y" %in% sampled)
})

test_that("Profile blocks are handled (statements inside skipped)", {
    code = '
    data {
        int N;
        array[N] real y;
    }
    parameters {
        real mu;
        real<lower=0> sigma;
    }
    model {
        mu ~ normal(0, 10);
        profile("likelihood") {
            y ~ normal(mu, sigma);
        }
    }'
    parsed = parse_model(code)
    sampled = vapply(parsed$samp, function(s) deparse(rlang::f_lhs(s)), "")
    expect_true("mu" %in% sampled)
    # profile block increases brace depth, so y ~ ... is skipped
    expect_false("y" %in% sampled)
})

test_that("Block comments and line comments are stripped", {
    code = "
    data {
        /* this is a block comment
           spanning multiple lines */
        int N;
        // real ghost_var;
    }
    model {
    }"
    parsed = parse_model(code)
    expect_equal(names(parsed$vars), "N")
    expect_true(! "ghost_var" %in% names(parsed$vars))
})

test_that("New constrained types are parsed", {
    code = "
    parameters {
        sum_to_zero_vector[5] beta;
        simplex[K] theta;
        ordered[3] cutpoints;
        cholesky_factor_corr[K] L;
    }
    model {
        beta ~ normal(0, 1);
        theta ~ dirichlet(alpha);
        cutpoints ~ normal(0, 5);
        L ~ lkj_corr_cholesky(2);
    }"
    parsed = parse_model(code)
    expect_equal(parsed$vars[c("beta", "theta", "cutpoints", "L")],
                 c(beta = "parameters", theta = "parameters",
                   cutpoints = "parameters", L = "parameters"))
})

test_that("Tokenizer handles Stan operators correctly", {
    tokens = tokenize_stan("target += normal_lpdf(y | mu, sigma);")
    expect_true("target" %in% tokens$value)
    expect_true("+=" %in% tokens$value)
    expect_true("|" %in% tokens$value)
})

test_that("Functions block is parsed without errors", {
    code = '
    functions {
        real my_fun(real x) {
            return x * 2;
        }
    }
    data {
        real y;
    }
    parameters {
        real mu;
    }
    model {
        mu ~ normal(0, 10);
        y ~ normal(mu, 1);
    }'
    parsed = parse_model(code)
    expect_equal(parsed$vars[c("y", "mu")], c(y = "data", mu = "parameters"))
    expect_length(parsed$samp, 2)
    expect_equal(parsed$samp[[2]], y ~ normal(mu, 1))
})

test_that("Truncated sampling statements are parsed (truncation stripped)", {
    code = "
    data {
        real<lower=0> y;
    }
    parameters {
        real mu;
        real<lower=0> sigma;
    }
    model {
        mu ~ normal(0, 10);
        sigma ~ exponential(1);
        y ~ normal(mu, sigma) T[0, ];
    }"
    parsed = parse_model(code)
    sampled = vapply(parsed$samp, function(s) deparse(rlang::f_lhs(s)), "")
    expect_true("y" %in% sampled)
    y_samp = parsed$samp[[which(sampled == "y")]]
    expect_equal(y_samp, y ~ normal(mu, sigma))
})

test_that("Multiple variables declared on separate lines in same block", {
    code = "
    parameters {
        real alpha;
        real beta;
        real<lower=0> sigma;
        vector[10] gamma;
    }
    model {
        alpha ~ normal(0, 1);
        beta ~ normal(0, 1);
        sigma ~ exponential(1);
        gamma ~ normal(0, 1);
    }"
    parsed = parse_model(code)
    expect_equal(sort(names(parsed$vars)),
                 sort(c("alpha", "beta", "sigma", "gamma")))
    expect_true(all(parsed$vars == "parameters"))
})

test_that("Transformed data and generated quantities variables are found", {
    code = "
    data {
        int N;
    }
    transformed data {
        real log_N = log(N);
    }
    parameters {
        real mu;
    }
    generated quantities {
        real y_rep = normal_rng(mu, 1);
    }
    model {
        mu ~ normal(0, 1);
    }"
    parsed = parse_model(code)
    expect_equal(parsed$vars["log_N"], c(log_N = "transformed data"))
    expect_equal(parsed$vars["y_rep"], c(y_rep = "generated quantities"))
})

test_that("Variable with assignment in declaration is parsed", {
    code = "
    transformed parameters {
        real<lower=0> disc = 1;
        vector[3] theta = mu + tau * eta;
    }
    parameters {
        real mu;
        real<lower=0> tau;
        vector[3] eta;
    }
    model {
        mu ~ normal(0, 1);
        tau ~ exponential(1);
        eta ~ std_normal();
    }"
    parsed = parse_model(code)
    expect_equal(parsed$vars["disc"], c(disc = "transformed parameters"))
    expect_equal(parsed$vars["theta"], c(theta = "transformed parameters"))
})

test_that("target += with _lupdf and _lupmf suffixes are parsed", {
    code = "
    data {
        real y;
    }
    parameters {
        real mu;
    }
    model {
        target += normal_lupdf(y | mu, 1);
    }"
    parsed = parse_model(code)
    expect_length(parsed$samp, 2)  # 1 explicit + 1 implicit uniform for mu
    expect_equal(parsed$samp[[1]], y ~ normal(mu, 1))
})

test_that("Mixed tilde and target += in same model", {
    code = "
    data {
        int N;
        array[N] real y;
    }
    parameters {
        real mu;
        real<lower=0> sigma;
    }
    model {
        mu ~ normal(0, 10);
        sigma ~ exponential(1);
        target += normal_lpdf(y | mu, sigma);
    }"
    parsed = parse_model(code)
    # 3 explicit statements, no implicit uniforms (all params have priors)
    expect_length(parsed$samp, 3)
    expect_equal(parsed$samp[[1]], mu ~ normal(0, 10))
    expect_equal(parsed$samp[[2]], sigma ~ exponential(1))
    expect_equal(parsed$samp[[3]], y ~ normal(mu, sigma))
})

#' @keywords internal
"_PACKAGE"

#' @import rlang
#' @import dplyr
NULL

# internal; to store shared package objects
pkg_env = new_environment()

.onLoad = function(libname, pkgname) {  # nocov start
    # create the Stan parser
    #tryCatch(get_parser(), error = function(e) {})

    utils::globalVariables(c("name", "pos", "value", ".y", ".y_ol", ".y_ou",
                             ".y_il", ".y_iu", ".y_med", "quantile", "median"))

    # Grab even more distributions from `extraDistr` if available
    distrs_onload()
}  # nocov end
#> NULL

#' Get Importance Resampling Indices From Weights
#'
#' Takes a vector of weights, or data frame or list containing sets of weights,
#' and resamples indices for use in later computation.
#'
#' @param x A vector of weights, a list of weight vectors, or a data frame of
#'   type \code{adjustr_weighted} containing a \code{.weights} list-column
#'   of weights.
#' @param frac A real number giving the fraction of draws to resample; the
#'   default, 1, resamples all draws. Smaller values should be used when
#'   \code{replace=FALSE}.
#' @param replace Whether sampling should be with replacement. When weights
#'   are extreme it may make sense to use \code{replace=FALSE}, but accuracy
#'   is not guaranteed in these cases.
#'
#' @return A vector, list, or data frame, depending of the type of \code{x},
#' containing the sampled indices. If any weights are \code{NA}, the indices
#' will also be \code{NA}.
#'
#' @examples
#' spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
#' adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)
#'
#' get_resampling_idxs(adjusted)
#' get_resampling_idxs(adjusted, frac=0.5, replace=FALSE)
#'
#' @export
get_resampling_idxs = function(x, frac=1, replace=TRUE) {
    if (frac < 0) stop("`frac` parameter must be nonnegative")
    get_idxs = function(w) {
        if (all(is.na(w))) return(NA_integer_)
        sample.int(length(w), size=round(frac*length(w)), replace=replace, prob=w)
    }

    if (inherits(x, "list")) {
        lapply(x, get_idxs)
    } else if (inherits(x, "adjustr_weighted")) {
        x$.idxs = lapply(x$.weights, get_idxs)
        x
    } else {
        get_idxs(x)
    }
}

#' Summarize Posterior Distributions Under Alternative Model Specifications
#'
#' Uses weights computed in \code{\link{adjust_weights}} to compute posterior
#' summary statistics. These statistics can be compared against their reference
#' values to quantify the sensitivity of the model to aspects of its
#' specification.
#'
#' @param .data An \code{adjustr_weighted} object.
#' @param ... Name-value pairs of expressions. The name of each argument will be
#'   the name of a new variable, and the value will be computed for the
#'   posterior distribution of eight alternative specification. For example,
#'   a value of \code{mean(theta)} will compute the posterior mean of
#'   \code{theta} for each alternative specification.
#'
#'   Also supported is the custom function \code{wasserstein}, which computes
#'   the Wasserstein-p distance between the posterior distribution of the
#'   provided expression under the new model and under the original model, with
#'   \code{p=1} the default. Lower the \code{spacing} parameter from the
#'   default of 0.005 to compute a finer (but slower) approximation.
#'
#'   The arguments in \code{...} are automatically quoted and evaluated in the
#'   context of \code{.data}. They support unquoting and splicing.
#' @param .resampling Whether to compute summary statistics by first resampling
#'  the data according to the weights. Defaults to \code{FALSE}, but will be
#'  used for any summary statistic that is not \code{mean}, \code{var} or
#'  \code{sd}.
#' @param .model_data Stan model data, if not provided in the earlier call to
#'   \code{\link{adjust_weights}}.
#'
#' @return An \code{adjustr_weighted} object, with the new columns specified in
#' \code{...} added.
#'
#' @seealso \code{\link{adjust_weights}}, \code{\link{spec_plot}}
#'
#' @examples
#' spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
#' adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)
#'
#' summarize(adjusted, mean(mu), var(mu))
#'
#' @rdname summarize.adjustr_weighted
#' @export
summarise.adjustr_weighted = function(.data, ..., .resampling=FALSE, .model_data=NULL) {
    stopifnot(is.adjustr_weighted(.data)) # just in case called manually
    args = enexprs(...)

    broadcast = function(x) {
        dims = c(dim(as.array(x)), iter)
        x = array(rep(x, iter), dim=dims)
        aperm(x, c(length(dims), 2:length(dims) - 1))
    }
    iter = attr(.data, "iter")
    if (!is_null(.model_data)) attr(.data, "data") = .model_data
    data = append(attr(.data, "draws"), lapply(attr(.data, "data"), broadcast))

    n_args = length(args)
    for (i in seq_along(args)) {
        name = names(args)[i]
        if (name == "") name = expr_name(args[[i]])

        call = args[[i]]
        if (!is_call(call)) {
            stop("Expressions must summarize posterior draws; `",
                 expr_text(call), "` has a different value for each draw.\n",
                 "  Try `mean(", expr_text(call), ")` or `sd(", expr_text(call), ")`.")
        }
        if (!.resampling && exists(call_name(call), funs_env)) {
            fun = funs_env[[call_name(call)]]
        } else {
            fun = function(x, ...) apply(x, 2, match.fun(call_name(call)), ...)
            .resampling = T
        }

        expr = expr_deparse(call_args(call)[[1]])
        expr = gsub("\\[(\\d)", "[,\\1", expr)
        expr = gsub("(?<![a-zA-Z0-9._])mean\\(", "rowMeans(", expr, perl=TRUE)
        expr = gsub("(?<![a-zA-Z0-9._])sum\\(", "rowSums(", expr, perl=TRUE)
        computed = as.array(eval_tidy(parse_expr(expr), data))
        if (length(dim(computed)) == 1) dim(computed) = c(dim(computed), 1)

        other_args = lapply(call_args(call)[-1], eval_tidy)
        if (!.resampling) {
            new_col = lapply(.data$.weights, function(w) {
                exec(fun, computed, w, !!!other_args)
            })
        } else {
            n_idx = max(min(5*iter, 20e3), iter)
            idxs = lapply(.data$.weights, function(w) sample.int(iter, n_idx, replace=TRUE, prob=w))
            new_col = lapply(idxs, function(idx) {
                comp = as.array(computed[idx,])
                if (length(dim(comp)) == 1) dim(comp) = c(dim(comp), 1)
                exec(fun, comp, !!!other_args)
            })
        }
        if (length(new_col[[1]]) == 1 && is.numeric(new_col[[1]]))
            new_col = as.numeric(new_col)
        .data[[name]] = new_col
    }

    .data
}
#' @rdname summarize.adjustr_weighted
#' @export
summarize.adjustr_weighted = summarise.adjustr_weighted

# Weighted ECDF
weighted.ecdf = function(samp, wgt=rep(1, length(samp))) {
    or = order(samp)
    y = cumsum(wgt[or])/sum(wgt)
    f = stats::stepfun(samp[or], c(0, y))
    class(f) = c("weighted.ecdf", "ecdf", class(f))
    attr(f, "call") = sys.call()
    f
}
#' @exportS3Method stats::quantile
quantile.weighted.ecdf = function(f, q) {
    x = environment(f)$x
    y = environment(f)$y
    vapply(q, function(q) {
        if (q == 0) return(x[1])
        if (q == 1) return(utils::tail(x, 1))
        idx = which(y > q)[1]
        if (idx == 1) return(x[1])
        stats::approx(y[idx-0:1], x[idx-0:1], q)$y
    }, numeric(1))
}

weighted.wasserstein = function(samp, wgt, p=1, spacing=0.005) {
    f = weighted.ecdf(samp, wgt)
    q = seq(0, 1, spacing)
    W = mean(abs(stats::quantile(samp, q, names=FALSE, type=4) - quantile.weighted.ecdf(f, q))^p)
    if (W < .Machine$double.eps) 0 else W^(1/p)
}

# Weighted summary functions that work on arrays
wtd_array_mean = function(arr, wgt) colSums(as.array(arr)*wgt) / sum(wgt)
wtd_array_var = function(arr, wgt)  wtd_array_mean((arr - wtd_array_mean(arr, wgt))^2, wgt)
wtd_array_sd = function(arr, wgt) sqrt(wtd_array_var(arr, wgt))
wtd_array_quantile = function(arr, wgt, probs=c(0.05, 0.25, 0.5, 0.75, 0.95)) {
    apply(arr, 2, function(x) quantile.weighted.ecdf(weighted.ecdf(x, wgt), probs))
}
wtd_array_median = function(arr, wgt) wtd_array_quantile(arr, wgt, 0.5)
wtd_array_wasserstein = function(arr, wgt, ...) {
    apply(arr, 2, function(x) weighted.wasserstein(x, wgt, ...))
}

funs_env = new_environment(list(
    mean = wtd_array_mean,
    var = wtd_array_var,
    sd = wtd_array_sd,
    quantile = wtd_array_quantile,
    median = wtd_array_median,
    wasserstein = wtd_array_wasserstein
))


#' Plot Posterior Quantities of Interest Under Alternative Model Specifications
#'
#' Uses weights computed in \code{\link{adjust_weights}} to plot posterior
#' quantities of interest versus specification parameters
#'
#' @param x An \code{adjustr_weighted} object.
#' @param by The x-axis variable, which is usually one of the specification
#'   parameters. Can be set to \code{1} if there is only one specification.
#'   Automatically quoted and evaluated in the context of \code{x}.
#' @param post The posterior quantity of interest, to be computed for each
#'   resampled draw of each specification. Should evaluate to a single number
#'   for each draw. Automatically quoted and evaluated in the context of \code{x}.
#' @param only_mean Whether to only plot the posterior mean. May be more stable.
#' @param ci_level The inner credible interval to plot. Central
#'   100*ci_level% intervals are computed from the quantiles of the resampled
#'   posterior draws.
#' @param outer_level The outer credible interval to plot.
#' @param ... Ignored.
#'
#' @return A \code{\link[ggplot2]{ggplot}} object which can be further
#'   customized with the \strong{ggplot2} package.
#'
#' @seealso \code{\link{adjust_weights}}, \code{\link{summarize.adjustr_weighted}}
#'
#' @examples
#' spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
#' adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)
#'
#' spec_plot(adjusted, df, mu, only_mean=TRUE)
#'
#' @export
spec_plot = function(x, by, post, only_mean=FALSE, ci_level=0.8,
                                 outer_level=0.95, ...) {
    if (!requireNamespace("ggplot2", quietly=TRUE)) { # nocov start
        stop("Package `ggplot2` must be installed to plot posterior quantities of interest.")
    } # nocov end
    if (ci_level > outer_level) stop("`ci_level` should be less than `outer_level`")

    post = enexpr(post)
    orig_row = filter(x, if_any(starts_with(".samp"), ~ . == "<original model>"))
    if (!only_mean) {
        outer = (1 - outer_level) / 2
        inner = (1 - ci_level) / 2
        q_probs = c(outer, inner, 0.5, 1-inner, 1-outer)
        sum_arg = quo(quantile(!!post, probs = !!q_probs))

        filter(x, if_any(starts_with(".samp"), ~ . != "<original model>")) %>%
            summarise.adjustr_weighted(.y = !!sum_arg) %>%
            rowwise() %>%
            mutate(.y_ol = .y[1],
                   .y_il = .y[2],
                   .y_med = .y[3],
                   .y_iu = .y[4],
                   .y_ou = .y[5]) %>%
        ggplot2::ggplot(ggplot2::aes({{ by }}, .y_med)) +
            ggplot2::geom_ribbon(ggplot2::aes(ymin=.y_ol, ymax=.y_ou), alpha=0.4) +
            ggplot2::geom_ribbon(ggplot2::aes(ymin=.y_il, ymax=.y_iu), alpha=0.5) +
            { if (nrow(orig_row) == 1)
                ggplot2::geom_hline(yintercept=summarise.adjustr_weighted(orig_row, .y = median(!!post))$`.y`,
                           lty="dashed")
            } +
            ggplot2::geom_line() +
            ggplot2::geom_point(size=3) +
            ggplot2::theme_minimal() +
            ggplot2::labs(y = expr_name(post))
    } else {
        filter(x, if_any(starts_with(".samp"), ~ . != "<original model>")) %>%
            summarise.adjustr_weighted(.y = mean(!!post)) %>%
        ggplot2::ggplot(ggplot2::aes({{ by }}, .y)) +
            { if (nrow(orig_row) == 1)
                ggplot2::geom_hline(yintercept=summarise.adjustr_weighted(orig_row, .y = mean(!!post))$`.y`,
                           lty="dashed")
            } +
            ggplot2::geom_line() +
            ggplot2::geom_point(size=3) +
            ggplot2::theme_minimal() +
            ggplot2::labs(y = expr_name(post))
    }
}

context("Summarizing and using computed weights")

test_that("Resampling indices can be made from a vector", {
    expect_equal(get_resampling_idxs(c(0, 0, 1)), c(3, 3, 3))
    expect_equal(get_resampling_idxs(c(0, 1, 0), frac=1/3), 2)
    expect_true(is.na(get_resampling_idxs(NA)))
})

test_that("Resampling indices can be made from a list", {
    expect_equal(get_resampling_idxs(list(c(0, 0, 1), c(0, 1, 0))),
                                     list(c(3, 3, 3), c(2, 2, 2)))
    expect_equal(get_resampling_idxs(list(NA, c(0, 1, 0))),
                                     list(NA_integer_, c(2, 2, 2)))
})

test_that("Resampling indices can be made from an adjustr_weighted object", {
    spec = make_spec(eta ~ student_t(7, 0, 1))
    obj = adjust_weights(spec, eightschools_m, keep_bad=TRUE)
    obj = get_resampling_idxs(obj)
    expect_is(obj$.idxs, "list")
    expect_length(obj$.idxs[[1]], 20)
})

test_that("Resampling indices fail with negative `frac`", {
    expect_error(get_resampling_idxs(c(0, 0, 1), frac=-0.5), "must be nonnegative")
})

test_that("Weighted array functions compute correctly", {
    y = as.array(1:5)
    dim(y) = c(dim(y), 1)
    wgt = c(1, 1, 2, 5, 1)
    wtd_mean = weighted.mean(y, wgt)

    expect_equal(wtd_array_mean(y, wgt), wtd_mean)
    expect_equal(wtd_array_var(y, wgt), weighted.mean((y - wtd_mean)^2, wgt))
    expect_equal(wtd_array_sd(y, wgt), sqrt(weighted.mean((y - wtd_mean)^2, wgt)))
    expect_equal(wtd_array_quantile(y, rep(1, 5), 0.2), 1)
    expect_equal(wtd_array_median(y, rep(1, 5)), 2.5)
})

test_that("Empty call to `summarize` should change nothing", {
    obj = tibble(.weights=list(c(1,1,1), c(1,1,4)))
    class(obj) = c("adjustr_weighted", class(obj))
    expect_identical(summarize(obj), obj)
})

test_that("Non-summary call to `summarize` should throw error", {
    obj = tibble(.weights=list(c(1,1,1), c(1,1,4)))
    attr(obj, "draws") = list(theta=matrix(c(3,5,7,1,1,1), ncol=2))
    attr(obj, "iter") = 3
    class(obj) = c("adjustr_weighted", class(obj))

    expect_error(summarize(obj, theta), "must summarize posterior draws")
})

test_that("Basic summaries are computed correctly", {
    obj = tibble(.weights=list(c(1,1,1), c(1,1,4)))
    attr(obj, "draws") = list(theta=matrix(c(3,5,7,1,1,1), ncol=2))
    attr(obj, "iter") = 3
    class(obj) = c("adjustr_weighted", class(obj))

    sum1 = summarize(obj, mean(theta[1]))
    expect_is(sum1, "adjustr_weighted")
    expect_equal(sum1$`mean(theta[1])`, 5:6)

    sum2 = summarize(obj, th = mean(theta))
    expect_is(sum2, "adjustr_weighted")
    expect_equal(sum2$th, list(c(5, 1), c(6, 1)))

    sum3 = summarize(obj, W = wasserstein(theta[1]))
    expect_is(sum3, "adjustr_weighted")
    expect_equal(sum3$W[1], 0)

    expect_error(summarise.adjustr_weighted(as_tibble(obj)), "is not TRUE")
})

test_that("`summarize` uses data correctly", {
    obj = tibble(.weights=list(c(1,1,1), c(1,1,4)))
    attr(obj, "draws") = list(eta=matrix(c(3,5,7), nrow=3, ncol=1))
    attr(obj, "iter") = 3
    class(obj) = c("adjustr_weighted", class(obj))
    model_d = list(theta=4)

    expect_equal(summarize(obj, th=mean(theta+eta), .model_data=model_d)$th, c(9,10))

    attr(obj, "data") = model_d
    expect_equal(summarize(obj, th=mean(theta+eta))$th, c(9,10))
})


test_that("Resampling-based summaries are computed correctly", {
    obj = tibble(.weights=list(c(1,0,0), c(0,0,1)))
    attr(obj, "draws") = list(theta=matrix(c(3,5,7), nrow=3, ncol=1))
    attr(obj, "iter") = 3
    class(obj) = c("adjustr_weighted", class(obj))

    sum1 = summarize(obj, th=mean(theta), .resampling=TRUE)
    expect_equal(sum1$th, c(3,7))

    sum2 = summarize(obj, th=quantile(theta, 0.05), .resampling=TRUE)
    expect_equal(sum2$th, c(3,7))
})


test_that("Weighted ECDF is computed correctly", {
    samp = 1:5
    wgt = rep(1, 5)
    f = weighted.ecdf(samp, wgt)
    expect_s3_class(f, "weighted.ecdf")
    expect_equal(f(0), 0)
    expect_equal(f(5), 1)
    expect_equal(f(3), 0.6)
})

test_that("Weighted ECDF with non-uniform weights", {
    samp = c(1, 2, 3)
    wgt = c(0.5, 0.25, 0.25)
    f = weighted.ecdf(samp, wgt)
    expect_equal(f(1), 0.5)
    expect_equal(f(3), 1)
})

test_that("Weighted ECDF quantile function at boundaries", {
    samp = 1:5
    f = weighted.ecdf(samp, rep(1, 5))
    expect_equal(quantile.weighted.ecdf(f, 0), 1)
    expect_equal(quantile.weighted.ecdf(f, 1), 5)
})

test_that("Wasserstein distance is zero for identical distributions", {
    samp = rnorm(100)
    w = weighted.wasserstein(samp, rep(1, 100), p = 1)
    expect_equal(w, 0)
})

test_that("Wasserstein distance is positive for different distributions", {
    set.seed(42)
    samp = rnorm(200)
    # weights that shift the distribution
    wgt = exp(samp)
    w = weighted.wasserstein(samp, wgt, p = 1)
    expect_gt(w, 0)
})

test_that("summarize computes multiple statistics at once", {
    obj = tibble(.weights = list(c(1, 1, 1), c(1, 1, 4)))
    attr(obj, "draws") = list(theta = matrix(c(3, 5, 7, 1, 1, 1), ncol = 2))
    attr(obj, "iter") = 3
    class(obj) = c("adjustr_weighted", class(obj))

    result = summarize(obj, m = mean(theta[1]), s = sd(theta[1]))
    expect_true("m" %in% names(result))
    expect_true("s" %in% names(result))
    expect_equal(result$m, c(5, 6))
})

test_that("summarize with var works correctly", {
    obj = tibble(.weights = list(c(1, 0, 0), c(0, 0, 1)))
    attr(obj, "draws") = list(theta = matrix(c(3, 5, 7), nrow = 3, ncol = 1))
    attr(obj, "iter") = 3
    class(obj) = c("adjustr_weighted", class(obj))

    result = summarize(obj, v = var(theta))
    # With all weight on one draw, variance should be 0
    expect_equal(result$v, c(0, 0))
})

test_that("summarize with median triggers resampling", {
    n = 1000
    vals = seq(1, 10, length.out = n)
    obj = tibble(.weights = list(
        c(rep(1, n/2), rep(0, n/2)),  # all weight on first half (median ~ 2.75)
        c(rep(0, n/2), rep(1, n/2))   # all weight on second half (median ~ 7.75)
    ))
    attr(obj, "draws") = list(theta = matrix(vals, nrow = n, ncol = 1))
    attr(obj, "iter") = n
    class(obj) = c("adjustr_weighted", class(obj))

    result = summarize(obj, m = median(theta))
    # true medians are 2.75 and 7.75; tolerance of 1.0 gives
    # essentially zero false positive rate with n=1000
    expect_lt(abs(result$m[1] - 2.75), 1.0)
    expect_lt(abs(result$m[2] - 7.75), 1.0)
})

test_that("Plotting function handles arguments correctly", {
    obj = tibble(.weights=list(c(1,0,0), c(0,0,1), c(1,1,1)),
                 .samp=c("y ~ normal(0, 1)", "y ~ normal(0, 2)", "<original model>"))
    attr(obj, "draws") = list(theta=matrix(c(3,5,7), nrow=3, ncol=1))
    attr(obj, "iter") = 3
    class(obj) = c("adjustr_weighted", class(obj))

    expect_is(spec_plot(obj, 1, theta), "ggplot")
    expect_is(spec_plot(obj, 1, theta, only_mean=TRUE), "ggplot")

    expect_error(spec_plot(obj, 1, theta, outer_level=0.4), "should be less than")
})

# Transpose a named list of equal-length vectors into a list of named lists.
# E.g., list(a=1:3, b=4:6) -> list(list(a=1,b=4), list(a=2,b=5), list(a=3,b=6))
list_transpose = function(x) {
    nms = names(x)
    n = length(x[[1]])
    lapply(seq_len(n), function(i) {
        out = lapply(x, `[[`, i)
        names(out) = nms
        out
    })
}

#' Set Up Model Adjustment Specifications
#'
#' Takes a set of new sampling statements, which can be parametrized by other
#' arguments, data frames, or lists, and creates an \code{adjustr_spec} object
#' suitable for use in \code{\link{adjust_weights}}.
#'
#' @param ... Model specification. Each argument can either be a formula,
#'   a named vector, data frames, or lists.
#'
#'   Formula arguments provide new sampling statements to replace their
#'   counterparts in the original Stan model. All such formulas must be of the
#'   form \code{variable ~ distribution(parameters)}, where \code{variable} and
#'   \code{parameters} are Stan data variables or parameters, or are provided by
#'   other arguments to this function (see below), and where \code{distribution}
#'   matches one of the univariate
#'   \href{https://mc-stan.org/docs/2_22/functions-reference/conventions-for-probability-functions.html}{Stan distributions}.
#'   Arithmetic expressions of parameters are also allowed, but care must be
#'   taken with multivariate parameter arguments.  Since specifications are
#'   passed as formulas, R's arithmetic operators are used, not Stan's. As a
#'   result, matrix and elementwise multipilcation in Stan sampling statements may
#'   not be interpreted correctly. Moving these computations out of sampling
#'   statements and into local variables will ensure correct results.
#'
#'   For named vector arguments, each entry of the vector will be substituted
#'   into the corresponding parameter in the sampling statements. For data
#'   frames, each entry in each column will be substituted into the corresponding
#'   parameter in the sampling statements.
#'
#'   List arguments are coerced to data frames. They can either be lists of named
#'   vectors, or lists of lists of single-element named vectors.
#'
#'   The lengths of all parameter arguments must be consistent.  Named vectors
#'   can have length 1 or must have length equal to the number of rows in all
#'   data frame arguments and the length of list arguments.
#'
#' @return An object of class \code{adjustr_spec}, which is essentially a list
#' with two elements: \code{samp}, which is a list of sampling formulas, and
#' \code{params}, which is a list of lists of parameters. Core
#' \link[=filter.adjustr_spec]{dplyr verbs} which don't involve grouping
#' (\code{\link[dplyr]{filter}}, \code{\link[dplyr]{arrange}},
#' \code{\link[dplyr]{mutate}}, \code{\link[dplyr]{select}},
#' \code{\link[dplyr]{rename}}, and \code{\link[dplyr]{slice}}) are
#' supported and operate on the underlying table of specification parameters.
#'
#' @param x An \code{adjustr_spec} object (for \code{print} and \code{length} methods).
#'
#' @seealso \code{\link{adjust_weights}}, \code{\link{summarize.adjustr_weighted}}, \code{\link{spec_plot}}
#'
#' @examples
#' make_spec(eta ~ cauchy(0, 1))
#'
#' make_spec(eta ~ student_t(df, 0, 1), df=1:10)
#'
#' params = tidyr::crossing(df=1:10, infl=c(1, 1.5, 2))
#' make_spec(eta ~ student_t(df, 0, 1),
#'           y ~ normal(theta, infl*sigma),
#'           params)
#'
#' @export
make_spec = function(...) {
    args = dots_list(..., .check_assign=TRUE)

    spec_samp = Filter(is_formula, args)
    if (length(spec_samp) == 0) warning("No sampling statements provided.")
    nms = names(args)
    spec_params = Map(function(value, name) {
        if (!is.numeric(name) && name != "") { # named arguments are preserved as is
            list2(!!name := value)
        } else if (is.data.frame(value)) {
            as.list(value)
        } else if (is_list(value)) {
            if (is_list(value[[1]])) {
                tryCatch(as.list(do.call(bind_rows, value)), error = function(e) {
                    stop("List-of-list arguments must be coercible to data frames.")
                })
            } else if (is_vector(value[[1]])) {
                tryCatch(do.call(bind_cols, value), error = function(e) {
                    stop("List-of-vector arguments must be coercible to data frames.")
                })
                value
            } else {
                stop("List arguments must be lists of lists or lists of ",
                    "vectors, and coercible to data frames.")
            }
        } else if (is_formula(value)) {
            NULL
        } else {
            stop("Arguments must be formulas, named vectors, data frames, ",
                "or lists. Use ?make_spec to see documentations.")
        }
    }, args, nms)
    spec_params = Filter(Negate(is.null), spec_params) # remove NULLs
    spec_params = do.call(c, unname(spec_params)) # flatten one level without name mangling
    spec_params = as_tibble(spec_params)

    if (any(is.na(spec_params)))
        stop("NAs found. Check input parameters and format.")

    spec_obj = structure(list(
        samp = spec_samp,
        params = if (nrow(spec_params) > 0)
                list_transpose(as.list(spec_params))
            else
                list(list())
    ),  class="adjustr_spec")
    spec_obj
}

# GENERIC FUNCTIONS for `adjustr_spec`
is.adjustr_spec = function(x) inherits(x, "adjustr_spec")
#' @rdname make_spec
#' @export
print.adjustr_spec = function(x, ...) {
    cat("Sampling specifications:\n")
    for (s in x$samp) print(s)
    if (length(x$params[[1]]) > 0) {
        cat("\nSpecification parameters:\n")
        df = as.data.frame(do.call(rbind, x$params))
        print(df, row.names=FALSE, max=15*ncol(df))
    }
    invisible(x)
}
#' @rdname make_spec
#' @export
length.adjustr_spec = function(x) length(x$params)

#' Convert an \code{adjustr_spec} Object Into a Data Frame
#'
#' Returns the data frame of specification parameters, with added columns of
#' the form \code{.samp_1}, \code{.samp_2}, ... for each sampling statement
#' (or just \code{.samp} if there is only one sampling statement).
#'
#' @param x the \code{adjustr_spec} object
#' @param ... additional arguments to underlying method
#'
#' @return A data frame with one row per specification and columns for each
#'   parameter, plus \code{.samp} (or \code{.samp_1}, \code{.samp_2}, ...)
#'   columns containing the sampling statement formulas.
#'
#' @export
as.data.frame.adjustr_spec = function(x, ...) {
    if (length(x$params[[1]]) == 0) {
        params_df = as.data.frame(matrix(nrow=1, ncol=0))
    } else {
        params_df = do.call(bind_rows, x$params) %>%
            as_tibble %>%
            as.data.frame
    }

    n_samp = length(x$samp)
    if (n_samp == 1) {
        params_df$.samp = format(x$samp[[1]])
    } else {
        for (i in 1:n_samp) {
            colname = paste0(".samp_", i)
            params_df[colname] = format(x$samp[[i]])
        }
    }

    params_df
}


#' \code{dplyr} Methods for \code{adjustr_spec} Objects
#'
#' Core \code{\link[dplyr]{dplyr}} verbs which don't involve grouping
#'  (\code{\link[dplyr]{filter}}, \code{\link[dplyr]{arrange}},
#'   \code{\link[dplyr]{mutate}}, \code{\link[dplyr]{select}},
#'   \code{\link[dplyr]{rename}}, and \code{\link[dplyr]{slice}}) are
#' implemented and operate on the underlying table of specification parameters.
#'
#' @param .data the \code{adjustr_spec} object
#' @param ... additional arguments to underlying method
#' @param .preserve as in \code{filter} and \code{slice}
#' @name dplyr.adjustr_spec
#'
#' @return A modified \code{adjustr_spec} object.
#'
#' @examples
#' spec = make_spec(eta ~ student_t(df, 0, 1), df=1:10)
#'
#' arrange(spec, desc(df))
#' slice(spec, 4:7)
#'
NULL
# dplyr generics
dplyr_handler = function(dplyr_func, x, ...) {
    if (length(x$params[[1]]) == 0) return(x)
    x$params = do.call(bind_rows, x$params) %>%
        as_tibble %>%
        dplyr_func(...) %>%
        as.list %>%
        list_transpose()
    x
}

#' @exportS3Method dplyr::filter
#' @rdname dplyr.adjustr_spec
filter.adjustr_spec = function(.data, ..., .preserve=FALSE) {
    dplyr_handler(dplyr::filter, .data, ..., .preserve=.preserve)
}
#' @rdname dplyr.adjustr_spec
#' @export
arrange.adjustr_spec = function(.data, ...) {
    dplyr_handler(dplyr::arrange, .data, ...)
}
#' @rdname dplyr.adjustr_spec
#' @export
rename.adjustr_spec = function(.data, ...) {
    dplyr_handler(dplyr::rename, .data, ...)
}
#' @rdname dplyr.adjustr_spec
#' @export
select.adjustr_spec = function(.data, ...) {
    dplyr_handler(dplyr::select, .data, ...)
}
#' @rdname dplyr.adjustr_spec
#' @export
slice.adjustr_spec = function(.data, ..., .preserve=FALSE) {
    dplyr_handler(dplyr::slice, .data, ..., .preserve=.preserve)
}

context("Specification creation")

test_that("Specifications can be created out of any number of formulas with no data", {
    spec1 = make_spec(y ~ normal(mu, sigma))
    expect_equal(spec1$params, list(list()))
    expect_length(spec1$samp, 1)
    expect_is(spec1$samp[[1]], "formula")

    spec2 = make_spec(y ~ normal(mu, sigma), sigma ~ gamma(alpha, beta))
    expect_equal(spec2$params, list(list()))
    expect_length(spec2$samp, 2)
    expect_is(spec2$samp[[1]], "formula")
    expect_is(spec2$samp[[2]], "formula")
})

test_that("Empty specifications generate warnigns", {
    expect_warning(make_spec(), "No sampling statements provided.")
    expect_warning(make_spec(df=1:5), "No sampling statements provided.")
})

test_that("Specifications can be created with named vectors", {
    spec = make_spec(y ~ std_normal(), df=1:5)
    expect_equal(spec$params, list(list(df=1), list(df=2), list(df=3), list(df=4), list(df=5)))
    expect_error(make_spec(y ~ std_normal(), 1:5), "named vector")
})

test_that("Specifications can be created with data frames", {
    dat = tibble(df=1:5)
    spec = make_spec(y ~ std_normal(), dat)
    expect_equal(spec$params, list(list(df=1), list(df=2), list(df=3), list(df=4), list(df=5)))
})

test_that("Specifications can be created with lists", {
    dat = tibble(df=1:5)
    ldat = list(list(df=1), list(df=2), list(df=3), list(df=4), list(df=5))
    expect_equal(make_spec(y ~ std_normal(), ldat)$params, ldat)
    expect_equal(make_spec(y ~ std_normal(), as.list(dat))$params, ldat)

    expect_error(make_spec(y ~ std_normal(), list(list(a=1:3), 7)),
                 "List-of-list arguments must be coercible to data frames")
    expect_error(make_spec(y ~ std_normal(), list(list(a=1:3), list(c=2:3))),
                 "NAs found. Check input parameters and format.")
    expect_error(make_spec(y ~ std_normal(), list(a=1:3, b=2:3)),
                 "List-of-vector arguments must be coercible to data frames")
    expect_error(make_spec(y ~ std_normal(), list(y~x)), "must be lists of lists")
})

test_that("Specifications generics work correctly", {
    spec = make_spec(y ~ std_normal(), df=1:5)

    expect_output(print(spec), "Specification parameters:\n df\n  1")
    expect_true(is.adjustr_spec(spec))
    expect_equal(length(spec), 5)

    spec2 = make_spec(y ~ std_normal(), df=1:5, mu=0:4)
    expect_equal(length(filter(spec2, df < 4)), 3)
    expect_equal(slice(arrange(spec2, desc(df)), 1)$params[[1]]$df, 5)
    expect_equal(names(rename(spec2, b=df)$params[[1]]), c("b", "mu"))
    expect_equal(length(select(spec2, df)$params[[1]]), 1)
    expect_equal(filter(make_spec(y ~ normal())), make_spec(y ~ normal()))
    expect_is(as.data.frame(spec2), "data.frame")
    expect_is(as.data.frame(make_spec(y ~ normal(), x ~ normal())), "data.frame")
})

test_that("Specifications with multiple parameters have correct structure", {
    spec = make_spec(eta ~ student_t(df, 0, scale), df = 1:3, scale = c(1, 2, 3))
    expect_length(spec, 3)
    expect_equal(spec$params[[1]]$df, 1)
    expect_equal(spec$params[[1]]$scale, 1)
    expect_equal(spec$params[[3]]$df, 3)
    expect_equal(spec$params[[3]]$scale, 3)
})

test_that("Specifications with crossing parameters via data frame", {
    params = tidyr::crossing(df = 1:3, scale = c(1, 2))
    spec = make_spec(eta ~ student_t(df, 0, scale), params)
    expect_length(spec, 6)
    expect_length(spec$samp, 1)
})

test_that("Specifications with multiple sampling statements", {
    spec = make_spec(eta ~ student_t(df, 0, 1),
                     y ~ normal(theta, infl * sigma),
                     df = 1:3, infl = c(1, 1.5, 2))
    expect_length(spec$samp, 2)
    expect_length(spec, 3)
})

test_that("as.data.frame for single vs multiple sampling statements", {
    spec1 = make_spec(y ~ normal(0, s), s = 1:3)
    df1 = as.data.frame(spec1)
    expect_true(".samp" %in% names(df1))
    expect_false(".samp_1" %in% names(df1))

    spec2 = make_spec(y ~ normal(0, s), x ~ normal(0, 1), s = 1:3)
    df2 = as.data.frame(spec2)
    expect_true(".samp_1" %in% names(df2))
    expect_true(".samp_2" %in% names(df2))
    expect_false(".samp" %in% names(df2))
})

test_that("Specifications with no parameters convert to data frame", {
    spec = make_spec(y ~ normal(0, 1))
    df = as.data.frame(spec)
    expect_is(df, "data.frame")
    expect_equal(nrow(df), 1)
})

test_that("slice preserves adjustr_spec class", {
    spec = make_spec(eta ~ student_t(df, 0, 1), df = 1:10)
    sliced = slice(spec, 3:5)
    expect_true(is.adjustr_spec(sliced))
    expect_length(sliced, 3)
    expect_equal(sliced$params[[1]]$df, 3)
})

test_that("NAs in parameters throw error", {
    expect_error(make_spec(y ~ normal(0, s), s = c(1, NA, 3)),
                 "NAs found")
})

#' Compute Pareto-smoothed Importance Weights for Alternative Model
#' Specifications
#'
#' Given a set of new sampling statements, which can be parametrized by a
#' data frame or list, compute Pareto-smoothed importance weights and attach
#' them to the specification object, for further calculation and plotting.
#'
#' This function does the bulk of the sensitivity analysis work.  It operates
#' by parsing the model code from the provided Stan object, extracting the
#' parameters and their sampling statements.  It then uses R
#' metaprogramming/tidy evaluation tools to flexibly evaluate the log density
#' for each draw and each sampling statement, under the original and alternative
#' specifications. From these, the function computes the overall importance
#' weight for each draw and performs Pareto-smoothed importance sampling.  All
#' of the work is performed in R, without recompiling or refitting the Stan
#' model.
#'
#' @param spec An object of class \code{adjustr_spec}, probably produced by
#'   \code{\link{make_spec}}, containing the new sampling sampling statements
#'   to replace their counterparts in the original Stan model, and the data,
#'   if any, by which these sampling statements are parametrized.
#' @param object A model object, either of type \code{\link[rstan]{stanfit}},
#'   \code{stanreg} (from \pkg{rstanarm}), \code{brmsfit} (from \pkg{brms}),
#'   or a list with two elements: \code{model} containing a
#'   \code{CmdStanModel}, and \code{fit} containing a
#'   \code{CmdStanMCMC} object (both from the \pkg{cmdstanr} package).
#' @param data The data that was used to fit the model in \code{object}.
#'   Required only if one of the new sampling specifications involves Stan data
#'   variables.
#' @param keep_bad When \code{FALSE} (the strongly recommended default),
#'   alternate specifications which deviate too much from the original
#'   posterior, and which as a result cannot be reliably estimated using
#'   importance sampling (i.e., if the Pareto shape parameter is larger than
#'   0.7), have their weights discarded—weights are set to \code{NA_real_}.
#' @param incl_orig When \code{TRUE}, include a row for the original
#'   model specification, with all weights equal. Can facilitate comparison
#'   and plotting later.
#'
#' @return A tibble, produced by converting the provided \code{specs} to a
#'   tibble (see \code{\link{as.data.frame.adjustr_spec}}), and adding columns
#'   \code{.weights}, containing vectors of weights for each draw, and
#'   \code{.pareto_k}, containing the diagnostic Pareto shape parameters. Values
#'   greater than 0.7 indicate that importance sampling is not reliable.
#'   If \code{incl_orig} is \code{TRUE}, a row is added for the original model
#'   specification. Weights can be extracted with the
#'   \code{\link{pull.adjustr_weighted}} method. The returned object also
#'   includes the model sample draws, in the \code{draws} attribute.
#'
#' @references
#' Vehtari, A., Simpson, D., Gelman, A., Yao, Y., & Gabry, J. (2024).
#' Pareto smoothed importance sampling.
#' \emph{Journal of Machine Learning Research}, 25(72), 1--58.
#' \doi{10.48550/arXiv.1507.02646}
#'
#' @seealso \code{\link{make_spec}}, \code{\link{summarize.adjustr_weighted}}, \code{\link{spec_plot}}
#'
#' @examples
#' spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
#' adjust_weights(spec, eightschools_m, keep_bad=TRUE)
#'
#' @export
adjust_weights = function(spec, object, data=NULL, keep_bad=FALSE, incl_orig=TRUE) {
    # CHECK ARGUMENTS
    if (is.null(data) & inherits(object, "brmsfit"))
        data = object$data
    object = get_fit_obj(object)
    stopifnot(is.adjustr_spec(spec))

    parsed = parse_model(object@stanmodel@model_code)

    # if no model data provided, we can only resample distributions of parameters
    if (is.null(data)) {
        samp_vars = vapply(parsed$samp, function(s) deparse(f_lhs(s)), "")
        prior_vars = parsed$vars[samp_vars] != "data"
        parsed$samp = parsed$samp[prior_vars]
        data = list()
    }

    matched_samp = match_sampling_stmts(spec$samp, parsed$samp)
    original_lp = calc_original_lp(object, matched_samp, parsed$vars, data)
    specs_lp = calc_specs_lp(object, spec$samp, parsed$vars, data, spec$params)

    # compute weights
    wgts = lapply(specs_lp, function(spec_lp) {
        lratio = spec_lp - original_lp
        dim(lratio) = c(dim(lratio), 1)
        r_eff = loo::relative_eff(as.array(exp(-lratio)))
        psis_wgt = suppressWarnings(loo::psis(lratio, r_eff=r_eff))
        pareto_k = loo::pareto_k_values(psis_wgt)
        if (all(psis_wgt$log_weights == psis_wgt$log_weights[1])) {
            warning("New specification equal to old specification.", call.=FALSE)
            pareto_k = -Inf
        }

        list(
            weights = loo::weights.importance_sampling(psis_wgt, log=FALSE),
            pareto_k = pareto_k
        )
    })

    adjust_obj = as_tibble(spec)
    class(adjust_obj) = c("adjustr_weighted", class(adjust_obj))
    adjust_obj$.weights = lapply(wgts, function(w) as.numeric(w$weights))
    adjust_obj$.pareto_k = vapply(wgts, function(w) w$pareto_k, numeric(1))
    if (!keep_bad)
        adjust_obj$.weights[adjust_obj$.pareto_k > 0.7] = list(NA_real_)
    attr(adjust_obj, "draws") = rstan::extract(object)
    attr(adjust_obj, "data") = data
    attr(adjust_obj, "iter") = object@sim$chains * (object@sim$iter - object@sim$warmup)
    if (incl_orig) {
        adjust_obj = bind_rows(adjust_obj, tibble(
            .weights=list(rep(1, attr(adjust_obj, "iter"))),
            .pareto_k = -Inf))
        samp_cols = grepl("\\.samp", names(adjust_obj))
        adjust_obj[nrow(adjust_obj), samp_cols] = "<original model>"
    }

    adjust_obj
}


# Generic methods
is.adjustr_weighted = function(x) inherits(x, "adjustr_weighted")
#' Extract Weights From an \code{adjustr_weighted} Object
#'
#' This function modifies the default behavior of \code{dplyr::pull} to extract
#' the \code{.weights} column.
#'
#' @param .data A table of data
#' @param var A variable, as in \code{\link[dplyr]{pull}}. The default returns
#'   the \code{.weights} column, and if there is only one row, it returns the
#'   first element of that column
#' @param name Ignored
#' @param ... Ignored
#'
#' @return A numeric vector of weights (if a single row) or a list of numeric
#'   vectors.
#'
#' @export
pull.adjustr_weighted = function(.data, var=".weights", name=NULL, ...) {
    var = tidyselect::vars_pull(names(.data), !!enquo(var))
    if (nrow(.data) == 1 && var == ".weights") {
        .data$.weights[[1]]
    } else {
        .data[[var]]
    }
}

#' Extract Model Sampling Statements From a Stan Model.
#'
#' Prints a list of sampling statements extracted from the \code{model} block of
#' a Stan program, with each labelled "parameter" or "data" depending on the
#' type of variable being sampled.
#'
#' @param object A \code{\link[rstan]{stanfit}} model object.
#'
#' @return Invisibly returns a list of sampling formulas.
#'
#' @examples
#' extract_samp_stmts(eightschools_m)
#' #> Sampling statements for model 2c8d1d8a30137533422c438f23b83428:
#' #>   parameter   eta ~ std_normal()
#' #>   data        y ~ normal(theta, sigma)
#'
#' @export
extract_samp_stmts = function(object) {
    object = get_fit_obj(object)

    parsed = parse_model(object@stanmodel@model_code)

    samp_vars = vapply(parsed$samp, function(s) rlang::expr_text(f_lhs(s)), "")
    samp_var_names = sub("\\[.+\\]", "", samp_vars)
    type = vapply(samp_var_names, function(var) {
        if (endsWith(parsed$vars[var], "data")) "data" else "parameter"
    }, "")
    print_order = order(type, samp_vars, decreasing=c(TRUE, FALSE), method="radix")

    cat(paste0("Sampling statements for model ", object@model_name, ":\n"))
    for (idx in print_order) {
        cat(sprintf("  %-9s   %s\n", type[idx], as.character(parsed$samp[idx])))
    }
    invisible(parsed$samp)
}

# Check that the model object is correct, and put it into a convenient format
get_fit_obj = function(object) {
    if (inherits(object, "stanfit")) {
        object
    } else if (inherits(object, "stanreg")) {
        object$stanfit
    } else if (inherits(object, "brmsfit")) {
        object$fit
    } else if (inherits(object, "list") && all(c("fit", "model") %in% names(object))
               && inherits(object$fit, "CmdStanMCMC") && inherits(object$model, "CmdStanModel")) {
        out = rstan::read_stan_csv(object$fit$output_files())
        out@stanmodel@model_code = paste0(object$model$code(), collapse="\n")
        out
    } else {
        stop("`object` must be of class `stanfit`, `stanreg`, `brmsfit`, or ",
             "a list with `CmdStanModel` and `CmdStanMCMC` objects.")
    }
}

context("Weight generation and model helpers")

setup({
    pkg_env$model_d = list(J = 8,
                           y = c(28,  8, -3,  7, -1,  1, 18, 12),
                           sigma = c(15, 10, 16, 11,  9, 11, 10, 18))
})

test_that("Identical specification gives warning", {
    spec = make_spec(y ~ normal(theta, sigma))
    expect_warning(adjust_weights(spec, eightschools_m, pkg_env$model_d), "equal to old")
})

test_that("High Pareto k values lead to discarded weights", {
    spec = make_spec(y ~ normal(theta, 1.1*sigma))
    obj = adjust_weights(spec, eightschools_m, pkg_env$model_d)
    expect_true(is.na(obj$.weights[[1]]))
})

test_that("Weights calculated correctly (normal/inflated)", {
    theta_draws = rstan::extract(eightschools_m, "theta", permuted=FALSE)
    y = pkg_env$model_d$y
    sigma = pkg_env$model_d$sigma

    ref_lp = apply(theta_draws, 1:2, function(theta) sum(dnorm(y, theta, sigma, log=TRUE)))
    new_lp = apply(theta_draws, 1:2, function(theta) sum(dnorm(y, theta, 1.1*sigma, log=TRUE)))
    lratio = new_lp - ref_lp
    dim(lratio) = c(dim(lratio), 1)
    r_eff = loo::relative_eff(as.array(exp(-lratio)))
    psis_wgt = suppressWarnings(loo::psis(lratio, r_eff=r_eff))
    pareto_k = loo::pareto_k_values(psis_wgt)
    weights = as.numeric(loo::weights.importance_sampling(psis_wgt, log=FALSE))

    spec = make_spec(y ~ normal(theta, 1.1*sigma))
    obj = adjust_weights(spec, eightschools_m, pkg_env$model_d, keep_bad=TRUE, incl_orig=FALSE)

    expect_s3_class(obj, "adjustr_weighted")
    expect_s3_class(obj, "tbl_df")
    expect_true(is.adjustr_weighted(obj))
    expect_true("draws" %in% names(attributes(obj)))
    expect_true("data" %in% names(attributes(obj)))
    expect_equal(weights, obj$.weights[[1]])
    expect_equal(pareto_k, obj$.pareto_k)
})

test_that("Weights calculated correctly (normal/student_t)", {
    theta_draws = rstan::extract(eightschools_m, "theta", permuted=FALSE)
    y = pkg_env$model_d$y
    sigma = pkg_env$model_d$sigma

    ref_lp = apply(theta_draws, 1:2, function(theta) sum(dnorm(y, theta, sigma, log=TRUE)))
    new_lp = apply(theta_draws, 1:2, function(theta) sum(dt((y-theta)/sigma, 6, log=TRUE)))
    lratio = new_lp - ref_lp
    dim(lratio) = c(dim(lratio), 1)
    r_eff = loo::relative_eff(exp(-lratio))
    psis_wgt = suppressWarnings(loo::psis(lratio, r_eff=r_eff))
    pareto_k = loo::pareto_k_values(psis_wgt)
    weights = as.numeric(loo::weights.importance_sampling(psis_wgt, log=FALSE))

    spec = make_spec(y ~ student_t(df, theta, sigma), df=5:6)
    obj = adjust_weights(spec, eightschools_m, pkg_env$model_d, keep_bad=TRUE, incl_orig=FALSE)

    expect_equal(weights, obj$.weights[[2]])
    expect_equal(pareto_k, obj$.pareto_k[2])
})

test_that("Weights calculated correctly (no data normal/student_t)", {
    eta_draws = rstan::extract(eightschools_m, "eta", permuted=FALSE)

    ref_lp = apply(eta_draws, 1:2, function(eta) sum(dnorm(eta, log=TRUE)))
    new_lp = apply(eta_draws, 1:2, function(eta) sum(dt(eta, 4, log=TRUE)))
    lratio = new_lp - ref_lp
    dim(lratio) = c(dim(lratio), 1)
    r_eff = loo::relative_eff(exp(-lratio))
    psis_wgt = suppressWarnings(loo::psis(lratio, r_eff=r_eff))
    pareto_k = loo::pareto_k_values(psis_wgt)
    weights = as.numeric(loo::weights.importance_sampling(psis_wgt, log=FALSE))

    spec = make_spec(eta ~ student_t(4, 0, 1))
    obj = adjust_weights(spec, eightschools_m, keep_bad=TRUE, incl_orig=FALSE)

    expect_equal(weights, obj$.weights[[1]])
    expect_equal(pareto_k, obj$.pareto_k)
})


test_that("Weights extracted correctly", {
    spec = make_spec(y ~ student_t(df, theta, sigma), df=5)
    obj = adjust_weights(spec, eightschools_m, pkg_env$model_d, keep_bad=TRUE, incl_orig=FALSE)
    pulled = pull(obj)

    expect_is(pulled, "numeric")
    expect_length(pulled, 20)

    spec2 = make_spec(y ~ student_t(df, theta, sigma), df=5:6)
    obj = adjust_weights(spec2, eightschools_m, pkg_env$model_d, keep_bad=TRUE, incl_orig=FALSE)
    pulled = pull(obj)

    expect_is(pulled, "list")
    expect_length(pulled, 2)
    expect_equal(vapply(pulled, length, 0L), c(20L, 20L))
})

test_that("Sampling statements printed correctly", {
    expect_output(extract_samp_stmts(eightschools_m),
"Sampling statements for model 2c8d1d8a30137533422c438f23b83428:
  parameter   eta ~ std_normal()
  parameter   mu ~ uniform(-1e+100, 1e+100)
  parameter   tau ~ uniform(-1e+100, 1e+100)
  data        y ~ normal(theta, sigma)", fixed=TRUE)
})

test_that("Fit objects extracted correctly", {
    obj = list(stanfit="stanreg", fit="brmsfit")

    class(obj) = "stanreg"
    expect_equal(get_fit_obj(obj), "stanreg")

    class(obj) = "brmsfit"
    expect_equal(get_fit_obj(obj), "brmsfit")

    class(obj) = "list"
    expect_error(get_fit_obj(obj), "must be of class")
})

test_that("adjust_weights includes original model row by default", {
    spec = make_spec(eta ~ student_t(4, 0, 1))
    obj = adjust_weights(spec, eightschools_m, keep_bad = TRUE, incl_orig = TRUE)
    expect_true(any(obj$.samp == "<original model>"))
    expect_equal(obj$.pareto_k[nrow(obj)], -Inf)
    # Original model weights should all be 1
    orig_wgts = obj$.weights[[nrow(obj)]]
    expect_true(all(orig_wgts == 1))
})

test_that("adjust_weights without original model row", {
    spec = make_spec(eta ~ student_t(4, 0, 1))
    obj = adjust_weights(spec, eightschools_m, keep_bad = TRUE, incl_orig = FALSE)
    expect_false(any(obj$.samp == "<original model>", na.rm = TRUE))
    expect_equal(nrow(obj), 1)
})

test_that("adjust_weights with multiple specifications", {
    spec = make_spec(eta ~ student_t(df, 0, 1), df = c(3, 5, 7))
    obj = adjust_weights(spec, eightschools_m, keep_bad = TRUE, incl_orig = FALSE)
    expect_equal(nrow(obj), 3)
    expect_length(obj$.weights, 3)
    expect_length(obj$.pareto_k, 3)
    # Each weight vector should have the right length
    expect_true(all(vapply(obj$.weights, length, 0L) == attr(obj, "iter")))
})

test_that("adjust_weights preserves spec parameters in output", {
    spec = make_spec(eta ~ student_t(df, 0, 1), df = c(3, 5, 7))
    obj = adjust_weights(spec, eightschools_m, keep_bad = TRUE, incl_orig = FALSE)
    expect_true("df" %in% names(obj))
    expect_equal(obj$df, c(3, 5, 7))
})

test_that("adjust_weights errors when data needed but not provided", {
    spec = make_spec(y ~ normal(theta, sigma))
    expect_error(adjust_weights(spec, eightschools_m))
})

#' Pre-fitted Eight Schools Model
#'
#' A small \code{\link[rstan]{stanfit}} object from the classic
#' "eight schools" hierarchical model (Rubin, 1981), included for use in
#' examples and tests.  The model was fit with 2 chains of 1000 iterations
#' (500 warmup) for a compact object size.
#'
#' @format A \code{\link[rstan]{stanfit}} object.
#' @keywords internal
#' @examples
#' extract_samp_stmts(eightschools_m)
"eightschools_m"

# Pass this function a density function.
# It will return a curried version which first takes the points at
# which to evaluate the density, and then takes the distribution parameters
make_dens = function(f) {
    function(x) {
        function(...) {
            f(x, ..., log=TRUE)
        }
    }
}

# Given a sampling formula and a `vars_data` object containing (1) model draws,
# (2) model input data (if applicable), and (3) prior specification data (if
# applicable), compute the log probability of the prior distribution for each
# MCMC draw of the parameter of interest
calc_lp = function(samp, vars_data) {
    # plug in RHS sampling distribution
    fn_name = as.character(f_rhs(samp)[[1]])
    distr = tryCatch(distr_env[[fn_name]],
                error=function(e)
                    stop("Distribution ", as.character(samp)[3], " not supported."))
    if (is.null(distr))
        stop("Distribution ", as.character(samp)[3], " not supported.")
    # plug in LHS, then RHS values. eval_tidy will throw error if no matches found
    distr = distr(eval_tidy(f_lhs(samp), vars_data))
    params = lapply(call_args(f_rhs(samp)), eval_tidy, vars_data)

    apply(exec(distr, !!!params), 1:2, sum)
}

# Given a list of sampling formulas, compile the necessary data from
# `object` MCMC draws and `data` model data
# NOTE: fills data first with MCMC data, so these values will override any
# passed in data
get_base_data = function(object, samps, parsed_vars, data, extra_names=NULL) {
    iter = object@sim$iter - object@sim$warmup
    chains = object@sim$chains
    reshape_data = function(x) {
        x = as.array(x)
        new_dim = c(iter, chains)
        new_x = array(rep(0, prod(new_dim)), dim=new_dim)
        aperm(apply(new_x, 1:2, function(y) x),
              c(length(dim(x)) + 1:2, 1:length(dim(x))))
    }

    lapply(samps, function(samp) {
        vars = get_stmt_vars(samp)
        # vars stored in MCMC draws
        vars_inmodel = intersect(vars, names(parsed_vars))
        vars_indraws = vars_inmodel[!endsWith(parsed_vars[vars_inmodel], "data")]
        vars_indata = vars_inmodel[endsWith(parsed_vars[vars_inmodel], "data")]
        # check data vars provided
        found = vars_indata %in% names(data)
        if (!all(found)) stop(paste(vars_indata[!found], collapse=", "), " not found")
        # combine draws and data
        draws_list = lapply(vars_indraws, function(v) rstan::extract(object, v, permuted=FALSE))
        names(draws_list) = vars_indraws
        data_list = lapply(vars_indata, function(v) reshape_data(data[[v]]))
        names(data_list) = vars_indata
        base_data = append(draws_list, data_list)
        # check all data found
        found = vars %in% c(names(base_data), extra_names)
        if (!all(found)) stop(paste(vars[!found], collapse=", "), " not found")
        base_data
    })
}

# Given a `samps` list of samp formula, compute the total log probability
# for each draw of the parameters of interest
calc_original_lp = function(object, samps, parsed_vars, data) {
    # figure out what data we need and calculate and sum lp
    base_data = get_base_data(object, samps, parsed_vars, data)
    Reduce(`+`, Map(calc_lp, samps, base_data))
}

# Given a `samps` list of samp formula, compute the total log probability for
# each draw of the parameters of interest, for each samp specification given
# in `specs`
calc_specs_lp = function(object, samps, parsed_vars, data, specs) {
    base_data = get_base_data(object, samps, parsed_vars, data, names(specs[[1]]))
    lapply(specs, function(spec) {
        Reduce(`+`, Map(function(s, bd) calc_lp(s, append(bd, spec)), samps, base_data))
    })
}


# Mapping of Stan distribution names to R functions
 distrs = list(
    bernoulli = function(x, p, ...) dbinom(x, 1, p, ...),
    bernoulli_logit = function(x, p, ...) dbinom(x, 1, plogis(p), ...),
    binomial = dbinom,
    binomial_logit = function(x, n, p, ...) dbinom(x, n, plogis(p), ...),
    hypergeometric = dhyper,
    poisson = dpois,
    poisson_log = function(x, alpha, ...) dpois(x, exp(alpha), ...),
    normal = dnorm,
    std_normal = dnorm,
    student_t = function(x, df, loc, scale, ...) dt((x - loc)/scale, df, ...),
    cauchy = dcauchy,
    laplace = function(x, mu, sigma, ...) dexp(abs(x - mu), 1/sigma, ...),
    logistic = dlogis,
    lognormal = dlnorm,
    chi_square = dchisq,
    exponential = dexp,
    gamma = function(x, alpha, beta, ...) dgamma(x, shape=alpha, rate=beta, ...),
    weibull = dweibull,
    beta = dbeta,
    beta_proportion = function(x, mu, k, ...) dbeta(x, mu*k, (1-mu)*k, ...),
    uniform = dunif
)
# Will be populated by distrs_onload() during .onLoad
distr_env = NULL
# Grab even more distributions from `extraDistr` if available (called from .onLoad())
distrs_onload = function() {
    if (requireNamespace("extraDistr", quietly=TRUE)) {
        distrs <<- append(distrs, list(
            beta_binomial = extraDistr::dbbinom,
            categorical = extraDistr::dcat,
            gumbel = extraDistr::dgumbel,
            inv_chi_square = extraDistr::dinvchisq,
            scaled_inv_chi_square = extraDistr::dinvchisq,
            inv_gamma = extraDistr::dinvgamma,
            frechet = function(x, lambda, sigma, ...)
                extraDistr::dfrechet(x, lambda, sigma=sigma, ...),
            rayleigh = extraDistr::drayleigh,
            pareto = extraDistr::dpareto
        ))
    } else {
        packageStartupMessage("`extraDistr` package not found. Install to access more ",
                "distributions, like inverse chi-square and beta-binomial.")
    }
    # Turn mapping into an environment suitable for metaprogramming,
    # and turn each density into its curried form (see `make_dens` above)
    distr_env <<- new_environment(lapply(distrs, make_dens))
}

context("Log probability calculations")

test_that("`make_dens` curries correctly", {
    f = make_dens(dexp)
    expect_error(f(1, 2), "unused argument") # right number of arguments
    test_val = runif(1, 0, 10)
    test_x = runif(1, 0.5, 2)
    expect_equal(f(test_x)(test_val), dexp(test_x, test_val, log=TRUE))
})

test_that("Constant parameter log probabilities are calculated correctly", {
    draws = matrix(-3:3, ncol=1)
    lp = calc_lp(y ~ std_normal(), list(y=draws))
    expect_equal(lp, matrix(dnorm(-3:3, log=TRUE), ncol=1))
})

test_that("Model parameter log probabilities are calculated correctly", {
    draws = matrix(-3:3, ncol=1)
    mus = -4:2
    sigmas = c(2, 1, 1, 1, 2, 1, 2)
    lp = calc_lp(y ~ normal(mu, sigma), list(y=draws, mu=mus, sigma=sigmas))
    expect_equal(lp, matrix(dnorm(-3:3, mus, sigmas, log=TRUE), ncol=1))
})

test_that("Data is assembled correctly", {
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    bd = get_base_data(eightschools_m, list(eta ~ student_t(df, 0, tau)),
                       parsed$vars, list(df=1:2), "df")

    expect_length(bd, 1)
    expect_named(bd[[1]], c("eta", "tau"), ignore.order=TRUE)
    expect_equal(dim(bd[[1]]$eta), c(10, 2, 8))
    expect_equal(dim(bd[[1]]$tau), c(10, 2, 1))
})

test_that("MCMC draws are preferred over provided data", {
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    bd = get_base_data(eightschools_m, list(eta ~ student_t(2, 0, tau)),
                       parsed$vars, list(tau=3))

    expect_length(bd, 1)
    expect_named(bd[[1]], c("eta", "tau"), ignore.order=TRUE)
    expect_equal(dim(bd[[1]]$eta), c(10, 2, 8))
    expect_equal(dim(bd[[1]]$tau), c(10, 2, 1))
})

test_that("Parameter-less specification data is correctly assembled", {
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    bd = get_base_data(eightschools_m, list(y ~ std_normal()), parsed$vars,
                       list(y=c(28,  8, -3,  7, -1,  1, 18, 12), J=8))

    expect_length(bd, 1)
    expect_named(bd[[1]], "y")
    expect_equal(dim(bd[[1]]$y), c(10, 2, 8))
})


test_that("Error thrown for missing data", {
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)

    expect_error(get_base_data(eightschools_m, list(eta ~ normal(gamma, sigma)),
                               parsed$vars, list()), "sigma not found")
    expect_error(get_base_data(eightschools_m, list(eta ~ normal(gamma, 2)),
                               parsed$vars, list()), "gamma not found")
})

test_that("Model log probability is correctly calculated", {
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    form = eta ~ normal(0, 1)
    draws = rstan::extract(eightschools_m, "eta", permuted=FALSE)
    exp_lp =  2*apply(dnorm(draws, 0, 1, log=TRUE), 1:2, sum)
    lp = calc_original_lp(eightschools_m, list(form, form), parsed$vars, list())
    expect_equal(exp_lp, lp)
})

test_that("Alternate specifications log probabilities are correctly calculated", {
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    form = eta ~ normal(0, s)
    draws = rstan::extract(eightschools_m, "eta", permuted=FALSE)
    exp_lp =  2*apply(dnorm(draws, 0, 1, log=TRUE), 1:2, sum)
    lp = calc_specs_lp(eightschools_m, list(form, form), parsed$vars, list(), list(list(s=1)))
    expect_equal(exp_lp, lp[[1]])
})

test_that("Multiple alternate specs produce correct number of results", {
    code = eightschools_m@stanmodel@model_code
    parsed = parse_model(code)
    form = eta ~ normal(0, s)
    specs = list(list(s = 1), list(s = 2), list(s = 3))
    lps = calc_specs_lp(eightschools_m, list(form), parsed$vars, list(), specs)
    expect_length(lps, 3)
    # All should be matrices of the same dimension
    expect_true(all(vapply(lps, function(x) identical(dim(x), dim(lps[[1]])), logical(1))))
})

test_that("Log probabilities for different distributions are finite", {
    draws = matrix(runif(10, 0.1, 5), ncol = 1)
    # Exponential
    lp_exp = calc_lp(y ~ exponential(1), list(y = draws))
    expect_true(all(is.finite(lp_exp)))
    expect_equal(lp_exp, matrix(dexp(draws, 1, log = TRUE), ncol = 1))

    # Gamma
    lp_gamma = calc_lp(y ~ gamma(2, 1), list(y = draws))
    expect_true(all(is.finite(lp_gamma)))
    expect_equal(lp_gamma, matrix(dgamma(draws, shape = 2, rate = 1, log = TRUE), ncol = 1))

    # Beta (values in (0,1))
    draws01 = matrix(runif(10, 0.01, 0.99), ncol = 1)
    lp_beta = calc_lp(y ~ beta(2, 3), list(y = draws01))
    expect_true(all(is.finite(lp_beta)))
    expect_equal(lp_beta, matrix(dbeta(draws01, 2, 3, log = TRUE), ncol = 1))
})

test_that("Unsupported distribution throws error", {
    draws = matrix(1:5, ncol = 1)
    expect_error(calc_lp(y ~ fake_distribution(1), list(y = draws)),
                 "not supported")
})

---
title: "Sensitivity Analysis of a Simple Hierarchical Model"
output: rmarkdown::html_vignette
bibliography: adjustr.bib
vignette: >
  %\VignetteIndexEntry{Sensitivity Analysis of a Simple Hierarchical Model}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(dplyr)
library(rstan)
library(adjustr)
load("eightschools_model.rda")
```

## Introduction

This vignette walks through the process of performing sensitivity
analysis using the `adjustr` package for the classic introductory
hierarchical model: the "eight schools" meta-analysis from Chapter 5
of @bda3.

We begin by specifying and fitting the model, which should be familiar
to most users of Stan. 
```{r eval=F}
library(dplyr)
library(rstan)
library(adjustr)

model_code = "
data {
    int<lower=0> J;         // number of schools
    real y[J];              // estimated treatment effects
    real<lower=0> sigma[J]; // standard error of effect estimates
}
parameters {
    real mu;                // population treatment effect
    real<lower=0> tau;      // standard deviation in treatment effects
    vector[J] eta;          // unscaled deviation from mu by school
}
transformed parameters {
    vector[J] theta = mu + tau * eta;        // school treatment effects
}
model {
    eta ~ std_normal();
    y ~ normal(theta, sigma);
}"

model_d = list(J = 8,
               y = c(28,  8, -3,  7, -1,  1, 18, 12),
               sigma = c(15, 10, 16, 11,  9, 11, 10, 18))
eightschools_m = stan(model_code=model_code, chains=2, data=model_d,
                      warmup=500, iter=1000)
```

We plot the original estimates for each of the eight schools.
```{r}
plot(eightschools_m, pars="theta")
```

The model partially pools information, pulling the school-level treatment effects
towards the overall mean.

It is natural to wonder how much these estimates depend on certain aspects of
our model.  The individual and school treatment effects are assumed to follow a
normal distribution, and we have used a uniform prior on the population
parameters `mu` and `tau`.

The basic __adjustr__ workflow is as follows:

1. Use `make_spec` to specify the set of alternative model specifications you'd
like to fit.

2. Use `adjust_weights` to calculate importance sampling weights which
approximate the posterior of each alternative specification.

3. Use `summarize` and `spec_plot` to examine posterior quantities of interest
for each alternative specification, in order to assess the sensitivity of the
underlying model.

## Basic Workflow Example

First suppose we want to examine the effect of our choice of uniform prior
on `mu` and `tau`.  We begin by specifying an alternative model in which
these parameters have more informative priors. This just requires
passing the `make_spec` function the new sampling statements we'd like to 
use. These replace any in the original model (`mu` and `tau` have implicit
improper uniform priors, since the original model does not have any sampling
statements for them).
```{r}
spec = make_spec(mu ~ normal(0, 20), tau ~ exponential(5))
print(spec)
```

Then we compute importance sampling weights to approximate the posterior under
this alternative model.
```{r include=F}
adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)
```
```{r eval=F}
adjusted = adjust_weights(spec, eightschools_m)
```

The `adjust_weights` function returns a data frame
containing a summary of the alternative model and a list-column named `.weights`
containing the importance weights. The last row of the table by default 
corresponds to the original model specification. The table also includes the
diagnostic Pareto *k*-value. When this value exceeds 0.7, importance sampling is
unreliable, and by default `adjust_weights` discards weights with a Pareto *k*
above 0.7 (the respective rows in `adjusted` are kept, but the `weights` column 
is set to `NA_real_`).
```{r}
print(adjusted)
```

Finally, we can examine how these alternative priors have changed our posterior
inference. We use `summarize` to calculate these under the alternative model.
```{r}
summarize(adjusted, mean(mu), var(mu))
```
We see that the more informative priors have pulled the posterior distribution
of `mu` towards zero and made it less variable.

## Multiple Alternative Specifications
What if instead we are concerned about our distributional assumption on the
school treatment effects? We could probe this assumption by fitting a series of
models where `eta` had a Student's *t* distribution, with varying degrees of
freedom. 

The `make_spec` function handles this easily.
```{r}
spec = make_spec(eta ~ student_t(df, 0, 1), df=1:10)
print(spec)
```
Notice how we have parameterized the alternative sampling statement with 
a variable `df`, and then provided the values `df` takes in another argument
to `make_spec`.

As before, we compute importance sampling weights to approximate the posterior
under these alternative models. Here, for the purposes of illustration,
we are using `keep_bad=TRUE` to compute weights even when the Pareto _k_ diagnostic
value is above 0.7. In practice, the alternative models should be completely
re-fit in Stan.
```{r}
adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)
```
Now, `adjusted` has ten rows, one for each alternative model.
```{r}
print(adjusted)
```

To examine the impact of these model changes, we can plot the posterior for 
a quantity of interest versus the degrees of freedom for the *t* distribution.
The package provides the `spec_plot` function which takes an x-axis specification
parameter and a y-axis posterior quantity (which must evaluate to a single
number per posterior draw). The dashed line shows the posterior median 
under the original model.
```{r}
spec_plot(adjusted, df, mu)
spec_plot(adjusted, df, theta[3])
```

It appears that changing the distribution of `eta`/`theta` from normal to
*t* has a small effect on posterior inferences (although, as noted above,
these inferences are unreliable as _k_ > 0.7).

By default, the function plots an inner 80\% credible interval and an outer
95\% credible interval, but these can be changed by the user.

We can also measure the distance between the new and original posterior
marginals by using the special `wasserstein()` function available in 
`summarize()`:
```{r}
summarize(adjusted, wasserstein(mu))
```
As we would expect, the 1-Wasserstein distance decreases as the degrees of
freedom increase. In general, we can compute the _p_-Wasserstein distance
by passing an extra `p` parameter to `wasserstein()`.


###

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjust_weights.R
\name{adjust_weights}
\alias{adjust_weights}
\title{Compute Pareto-smoothed Importance Weights for Alternative Model
Specifications}
\usage{
adjust_weights(spec, object, data = NULL, keep_bad = FALSE, incl_orig = TRUE)
}
\arguments{
\item{spec}{An object of class \code{adjustr_spec}, probably produced by
\code{\link{make_spec}}, containing the new sampling sampling statements
to replace their counterparts in the original Stan model, and the data,
if any, by which these sampling statements are parametrized.}

\item{object}{A model object, either of type \code{\link[rstan]{stanfit}},
\code{stanreg} (from \pkg{rstanarm}), \code{brmsfit} (from \pkg{brms}),
or a list with two elements: \code{model} containing a
\code{CmdStanModel}, and \code{fit} containing a
\code{CmdStanMCMC} object (both from the \pkg{cmdstanr} package).}

\item{data}{The data that was used to fit the model in \code{object}.
Required only if one of the new sampling specifications involves Stan data
variables.}

\item{keep_bad}{When \code{FALSE} (the strongly recommended default),
alternate specifications which deviate too much from the original
posterior, and which as a result cannot be reliably estimated using
importance sampling (i.e., if the Pareto shape parameter is larger than
0.7), have their weights discarded—weights are set to \code{NA_real_}.}

\item{incl_orig}{When \code{TRUE}, include a row for the original
model specification, with all weights equal. Can facilitate comparison
and plotting later.}
}
\value{
A tibble, produced by converting the provided \code{specs} to a
  tibble (see \code{\link{as.data.frame.adjustr_spec}}), and adding columns
  \code{.weights}, containing vectors of weights for each draw, and
  \code{.pareto_k}, containing the diagnostic Pareto shape parameters. Values
  greater than 0.7 indicate that importance sampling is not reliable.
  If \code{incl_orig} is \code{TRUE}, a row is added for the original model
  specification. Weights can be extracted with the
  \code{\link{pull.adjustr_weighted}} method. The returned object also
  includes the model sample draws, in the \code{draws} attribute.
}
\description{
Given a set of new sampling statements, which can be parametrized by a
data frame or list, compute Pareto-smoothed importance weights and attach
them to the specification object, for further calculation and plotting.
}
\details{
This function does the bulk of the sensitivity analysis work.  It operates
by parsing the model code from the provided Stan object, extracting the
parameters and their sampling statements.  It then uses R
metaprogramming/tidy evaluation tools to flexibly evaluate the log density
for each draw and each sampling statement, under the original and alternative
specifications. From these, the function computes the overall importance
weight for each draw and performs Pareto-smoothed importance sampling.  All
of the work is performed in R, without recompiling or refitting the Stan
model.
}
\examples{
spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
adjust_weights(spec, eightschools_m, keep_bad=TRUE)

}
\references{
Vehtari, A., Simpson, D., Gelman, A., Yao, Y., & Gabry, J. (2024).
Pareto smoothed importance sampling.
\emph{Journal of Machine Learning Research}, 25(72), 1--58.
\doi{10.48550/arXiv.1507.02646}
}
\seealso{
\code{\link{make_spec}}, \code{\link{summarize.adjustr_weighted}}, \code{\link{spec_plot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjust_weights.R
\name{pull.adjustr_weighted}
\alias{pull.adjustr_weighted}
\title{Extract Weights From an \code{adjustr_weighted} Object}
\usage{
\method{pull}{adjustr_weighted}(.data, var = ".weights", name = NULL, ...)
}
\arguments{
\item{.data}{A table of data}

\item{var}{A variable, as in \code{\link[dplyr]{pull}}. The default returns
the \code{.weights} column, and if there is only one row, it returns the
first element of that column}

\item{name}{Ignored}

\item{...}{Ignored}
}
\value{
A numeric vector of weights (if a single row) or a list of numeric
  vectors.
}
\description{
This function modifies the default behavior of \code{dplyr::pull} to extract
the \code{.weights} column.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_weights.R
\name{summarise.adjustr_weighted}
\alias{summarise.adjustr_weighted}
\alias{summarize.adjustr_weighted}
\title{Summarize Posterior Distributions Under Alternative Model Specifications}
\usage{
\method{summarise}{adjustr_weighted}(.data, ..., .resampling = FALSE, .model_data = NULL)

\method{summarize}{adjustr_weighted}(.data, ..., .resampling = FALSE, .model_data = NULL)
}
\arguments{
\item{.data}{An \code{adjustr_weighted} object.}

\item{...}{Name-value pairs of expressions. The name of each argument will be
  the name of a new variable, and the value will be computed for the
  posterior distribution of eight alternative specification. For example,
  a value of \code{mean(theta)} will compute the posterior mean of
  \code{theta} for each alternative specification.

  Also supported is the custom function \code{wasserstein}, which computes
  the Wasserstein-p distance between the posterior distribution of the
  provided expression under the new model and under the original model, with
  \code{p=1} the default. Lower the \code{spacing} parameter from the
  default of 0.005 to compute a finer (but slower) approximation.

  The arguments in \code{...} are automatically quoted and evaluated in the
  context of \code{.data}. They support unquoting and splicing.}

\item{.resampling}{Whether to compute summary statistics by first resampling
the data according to the weights. Defaults to \code{FALSE}, but will be
used for any summary statistic that is not \code{mean}, \code{var} or
\code{sd}.}

\item{.model_data}{Stan model data, if not provided in the earlier call to
\code{\link{adjust_weights}}.}
}
\value{
An \code{adjustr_weighted} object, with the new columns specified in
\code{...} added.
}
\description{
Uses weights computed in \code{\link{adjust_weights}} to compute posterior
summary statistics. These statistics can be compared against their reference
values to quantify the sensitivity of the model to aspects of its
specification.
}
\examples{
spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)

summarize(adjusted, mean(mu), var(mu))

}
\seealso{
\code{\link{adjust_weights}}, \code{\link{spec_plot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_spec.R
\name{dplyr.adjustr_spec}
\alias{dplyr.adjustr_spec}
\alias{filter.adjustr_spec}
\alias{arrange.adjustr_spec}
\alias{rename.adjustr_spec}
\alias{select.adjustr_spec}
\alias{slice.adjustr_spec}
\title{\code{dplyr} Methods for \code{adjustr_spec} Objects}
\usage{
\method{filter}{adjustr_spec}(.data, ..., .preserve = FALSE)

\method{arrange}{adjustr_spec}(.data, ...)

\method{rename}{adjustr_spec}(.data, ...)

\method{select}{adjustr_spec}(.data, ...)

\method{slice}{adjustr_spec}(.data, ..., .preserve = FALSE)
}
\arguments{
\item{.data}{the \code{adjustr_spec} object}

\item{...}{additional arguments to underlying method}

\item{.preserve}{as in \code{filter} and \code{slice}}
}
\value{
A modified \code{adjustr_spec} object.
}
\description{
Core \code{\link[dplyr]{dplyr}} verbs which don't involve grouping
 (\code{\link[dplyr]{filter}}, \code{\link[dplyr]{arrange}},
  \code{\link[dplyr]{mutate}}, \code{\link[dplyr]{select}},
  \code{\link[dplyr]{rename}}, and \code{\link[dplyr]{slice}}) are
implemented and operate on the underlying table of specification parameters.
}
\examples{
spec = make_spec(eta ~ student_t(df, 0, 1), df=1:10)

arrange(spec, desc(df))
slice(spec, 4:7)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{eightschools_m}
\alias{eightschools_m}
\title{Pre-fitted Eight Schools Model}
\format{
A \code{\link[rstan]{stanfit}} object.
}
\usage{
eightschools_m
}
\description{
A small \code{\link[rstan]{stanfit}} object from the classic
"eight schools" hierarchical model (Rubin, 1981), included for use in
examples and tests.  The model was fit with 2 chains of 1000 iterations
(500 warmup) for a compact object size.
}
\examples{
extract_samp_stmts(eightschools_m)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_weights.R
\name{spec_plot}
\alias{spec_plot}
\title{Plot Posterior Quantities of Interest Under Alternative Model Specifications}
\usage{
spec_plot(
  x,
  by,
  post,
  only_mean = FALSE,
  ci_level = 0.8,
  outer_level = 0.95,
  ...
)
}
\arguments{
\item{x}{An \code{adjustr_weighted} object.}

\item{by}{The x-axis variable, which is usually one of the specification
parameters. Can be set to \code{1} if there is only one specification.
Automatically quoted and evaluated in the context of \code{x}.}

\item{post}{The posterior quantity of interest, to be computed for each
resampled draw of each specification. Should evaluate to a single number
for each draw. Automatically quoted and evaluated in the context of \code{x}.}

\item{only_mean}{Whether to only plot the posterior mean. May be more stable.}

\item{ci_level}{The inner credible interval to plot. Central
100*ci_level% intervals are computed from the quantiles of the resampled
posterior draws.}

\item{outer_level}{The outer credible interval to plot.}

\item{...}{Ignored.}
}
\value{
A \code{\link[ggplot2]{ggplot}} object which can be further
  customized with the \strong{ggplot2} package.
}
\description{
Uses weights computed in \code{\link{adjust_weights}} to plot posterior
quantities of interest versus specification parameters
}
\examples{
spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)

spec_plot(adjusted, df, mu, only_mean=TRUE)

}
\seealso{
\code{\link{adjust_weights}}, \code{\link{summarize.adjustr_weighted}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use_weights.R
\name{get_resampling_idxs}
\alias{get_resampling_idxs}
\title{Get Importance Resampling Indices From Weights}
\usage{
get_resampling_idxs(x, frac = 1, replace = TRUE)
}
\arguments{
\item{x}{A vector of weights, a list of weight vectors, or a data frame of
type \code{adjustr_weighted} containing a \code{.weights} list-column
of weights.}

\item{frac}{A real number giving the fraction of draws to resample; the
default, 1, resamples all draws. Smaller values should be used when
\code{replace=FALSE}.}

\item{replace}{Whether sampling should be with replacement. When weights
are extreme it may make sense to use \code{replace=FALSE}, but accuracy
is not guaranteed in these cases.}
}
\value{
A vector, list, or data frame, depending of the type of \code{x},
containing the sampled indices. If any weights are \code{NA}, the indices
will also be \code{NA}.
}
\description{
Takes a vector of weights, or data frame or list containing sets of weights,
and resamples indices for use in later computation.
}
\examples{
spec = make_spec(eta ~ student_t(df, 0, 1), df=4:10)
adjusted = adjust_weights(spec, eightschools_m, keep_bad=TRUE)

get_resampling_idxs(adjusted)
get_resampling_idxs(adjusted, frac=0.5, replace=FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjustr-package.R
\docType{package}
\name{adjustr-package}
\alias{adjustr}
\alias{adjustr-package}
\title{adjustr: Stan Model Adjustments and Sensitivity Analyses using Importance Sampling}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Assess the sensitivity of a Bayesian model (fitted using 'Stan' via 'rstan', 'brms', or 'cmdstanr') to the specification of its likelihood and priors. Users provide a series of alternate sampling specifications, and the package uses Pareto-smoothed importance sampling (PSIS) to estimate posterior quantities of interest under each specification, without needing to refit the model. Methods are based on Vehtari, Simpson, Gelman, Yao, and Gabry (2024) \doi{10.48550/arXiv.1507.02646}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://corymccartan.com/adjustr/}
  \item Report bugs at \url{https://github.com/CoryMcCartan/adjustr/issues}
}

}
\author{
\strong{Maintainer}: Cory McCartan \email{mccartan@psu.edu} [copyright holder]

Authors:
\itemize{
  \item Cory McCartan \email{mccartan@psu.edu} [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_spec.R
\name{as.data.frame.adjustr_spec}
\alias{as.data.frame.adjustr_spec}
\title{Convert an \code{adjustr_spec} Object Into a Data Frame}
\usage{
\method{as.data.frame}{adjustr_spec}(x, ...)
}
\arguments{
\item{x}{the \code{adjustr_spec} object}

\item{...}{additional arguments to underlying method}
}
\value{
A data frame with one row per specification and columns for each
  parameter, plus \code{.samp} (or \code{.samp_1}, \code{.samp_2}, ...)
  columns containing the sampling statement formulas.
}
\description{
Returns the data frame of specification parameters, with added columns of
the form \code{.samp_1}, \code{.samp_2}, ... for each sampling statement
(or just \code{.samp} if there is only one sampling statement).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjust_weights.R
\name{extract_samp_stmts}
\alias{extract_samp_stmts}
\title{Extract Model Sampling Statements From a Stan Model.}
\usage{
extract_samp_stmts(object)
}
\arguments{
\item{object}{A \code{\link[rstan]{stanfit}} model object.}
}
\value{
Invisibly returns a list of sampling formulas.
}
\description{
Prints a list of sampling statements extracted from the \code{model} block of
a Stan program, with each labelled "parameter" or "data" depending on the
type of variable being sampled.
}
\examples{
extract_samp_stmts(eightschools_m)
#> Sampling statements for model 2c8d1d8a30137533422c438f23b83428:
#>   parameter   eta ~ std_normal()
#>   data        y ~ normal(theta, sigma)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_spec.R
\name{make_spec}
\alias{make_spec}
\alias{print.adjustr_spec}
\alias{length.adjustr_spec}
\title{Set Up Model Adjustment Specifications}
\usage{
make_spec(...)

\method{print}{adjustr_spec}(x, ...)

\method{length}{adjustr_spec}(x)
}
\arguments{
\item{...}{Model specification. Each argument can either be a formula,
  a named vector, data frames, or lists.

  Formula arguments provide new sampling statements to replace their
  counterparts in the original Stan model. All such formulas must be of the
  form \code{variable ~ distribution(parameters)}, where \code{variable} and
  \code{parameters} are Stan data variables or parameters, or are provided by
  other arguments to this function (see below), and where \code{distribution}
  matches one of the univariate
  \href{https://mc-stan.org/docs/2_22/functions-reference/conventions-for-probability-functions.html}{Stan distributions}.
  Arithmetic expressions of parameters are also allowed, but care must be
  taken with multivariate parameter arguments.  Since specifications are
  passed as formulas, R's arithmetic operators are used, not Stan's. As a
  result, matrix and elementwise multipilcation in Stan sampling statements may
  not be interpreted correctly. Moving these computations out of sampling
  statements and into local variables will ensure correct results.

  For named vector arguments, each entry of the vector will be substituted
  into the corresponding parameter in the sampling statements. For data
  frames, each entry in each column will be substituted into the corresponding
  parameter in the sampling statements.

  List arguments are coerced to data frames. They can either be lists of named
  vectors, or lists of lists of single-element named vectors.

  The lengths of all parameter arguments must be consistent.  Named vectors
  can have length 1 or must have length equal to the number of rows in all
  data frame arguments and the length of list arguments.}

\item{x}{An \code{adjustr_spec} object (for \code{print} and \code{length} methods).}
}
\value{
An object of class \code{adjustr_spec}, which is essentially a list
with two elements: \code{samp}, which is a list of sampling formulas, and
\code{params}, which is a list of lists of parameters. Core
\link[=filter.adjustr_spec]{dplyr verbs} which don't involve grouping
(\code{\link[dplyr]{filter}}, \code{\link[dplyr]{arrange}},
\code{\link[dplyr]{mutate}}, \code{\link[dplyr]{select}},
\code{\link[dplyr]{rename}}, and \code{\link[dplyr]{slice}}) are
supported and operate on the underlying table of specification parameters.
}
\description{
Takes a set of new sampling statements, which can be parametrized by other
arguments, data frames, or lists, and creates an \code{adjustr_spec} object
suitable for use in \code{\link{adjust_weights}}.
}
\examples{
make_spec(eta ~ cauchy(0, 1))

make_spec(eta ~ student_t(df, 0, 1), df=1:10)

params = tidyr::crossing(df=1:10, infl=c(1, 1.5, 2))
make_spec(eta ~ student_t(df, 0, 1),
          y ~ normal(theta, infl*sigma),
          params)

}
\seealso{
\code{\link{adjust_weights}}, \code{\link{summarize.adjustr_weighted}}, \code{\link{spec_plot}}
}
